#!/usr/bin/env python

# Experimental Gimp plugin to replicate a path along another path.
#
# History:
# v0.1: 2021-12-28: Plugin "Replicate path along path". Experimental.
# v0.2: 2022-01-10: Developing
# v0.3: 2022-01-17: Tapering
# v0.4: 2022-01-22: Developing (overshoots etc)
# v0.5: 2022-01-28: Better tapering
# v0.6: 2022-01-30: Bug fix: case ncopies=0
# v0.7: 2022-02-02: Making the plugin faster
# v0.8: 2022-04-28: Bug fix

# (c) Markku Koppinen 2021
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published
#   by the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This very file is the complete source code to the program.
#
#   If you make and redistribute changes to this code, please mark it
#   in reasonable ways as different from the original version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The GPL v3 licence is available at: https://www.gnu.org/licenses/gpl-3..en.html'


from __future__ import division, print_function
from gimpfu import *
from math import *
from cmath import exp as cexp
from copy import deepcopy

#==============================================================
#             class BCurve       
#==============================================================

# Note: The plane is treated as the complex plane.

class BCurve(object):
    """Data structures to handle Gimp vectors objects by means of individual
    Bezier arcs.
    By a "Bezier arc" we mean here a Bezier curve determined by 4 control points
    p0,p1,p2,p3.
    This class contains data structures needed to do computations on
    a Bezier curve which are done arc by arc.
    To this end, Gimp data structures
    - gimp.Vectors
    - gimp.VectorsBezierStroke
    are here represented as data structures (classes)
    - GimpVectors
    - GimpStroke
    - BezierCurve
    - BezierArc
    - ControlPoint = complex
    with the aim that
    - each control point [x,y] is represented as an instance of ControlPoint;
    - each Bezier arc (four control points, see above) is represented as an instance
      of BezierArc.
    """
    
    ControlPoint = complex
    
    class GimpVectors(object):
        """Essentially same data as in a gimp.Vectors object except that
        strokes are instances of GimpStroke rather than
        gimp.VectorsBezierStroke.
        Attributes:
        - stroke_list: [GimpStroke]
        - name:        string
        """
        def __init__(self,
                     stroke_list,            # [GimpStroke]
                     name='GimpVectors'      # string
                     ):
            self.stroke_list = stroke_list
            self.name = name
        def __str__(self):
            s = 'GimpVectors '+self.name
            count = 0
            for stroke in self.stroke_list:
                s += '\n  Stroke '+str(count)+': GimpStroke'
                s += '\n    '+str(stroke)
                count += 1
            return s
        def gv2vectors_object(self, image, name=None):
            """Conversion GimpVectors -> gimp.Vectors
            (The converse: vectors_object2gv)
            """
            def cp_list2xy_list(cp_list):
                xy = []
                for cp in cp_list:
                    xy += [cp.real, cp.imag]
                return xy
            if name is None:
                name = self.name
            vectors = pdb.gimp_vectors_new(image, name)
            for stroke in self.stroke_list:
                closed = stroke.closed
                xy_list = cp_list2xy_list(stroke.cp_list)
                stroke_id = pdb.gimp_vectors_stroke_new_from_points(
                             vectors, 0, len(xy_list), xy_list, closed)
            return vectors
    
    class GimpStroke(object):
        """Essentially same data as in a gimp.VectorsBezierStroke.points,
        except that control points list [x,y,x,y,...] is arranged as
        a list [ControlPoint] by joining successive x,y.
        Attributes:
        - cp_list:     [ControlPoint]
        - closed:      boolean
        - stroke_name: string
        """
        def __init__(self,
                     cp_list,                   # [ControlPoint]
                     closed,                    # boolean
                     stroke_name = 'GimpStroke' # string
                     ):
            self.stroke_name = stroke_name
            self.cp_list = cp_list
            self.closed = closed
        def __str__(self):
            s = self.stroke_name + '; Control points:'
            for cp in self.cp_list:
                s += '\n    '+str(cp)
            s += '\n    closed: '+str(self.closed)
            return s
        def gs2bc(self):
            """Conversion GimpStroke -> BezierCurve
            """
            head = self.cp_list[0]
            tail = self.cp_list[-1]
            inner = self.cp_list[1:-1]
            if len(inner) > 1: # Must be divisible by 4
                ba_list = []
                count = 0
                for i in range(0,len(inner)-1,3):
                    ba_list.append(BCurve.BezierArc(inner[i:i+4], 'arc '+str(count)))
                    count += 1
            elif len(inner) == 1: # Stroke has only one anchor
                ba_list = [BCurve.BezierArc([inner[0]], 'arc 0')]
            else:
                raise Exception("BCurve.GimpStroke.gs2bc: No anchors in stroke?")
            return BCurve.BezierCurve(
                               bezier_arcs = ba_list,
                               head_handle = head,
                               tail_handle = tail,
                               closed = self.closed,
                               curve_name = self.stroke_name
                               )
    
    class BezierCurve(object):
        """BezierCurve is a list of butting Bezier arcs with some extra data.
        Attributes:
        - curve_name:  string
        - bezier_arcs: [BezierArc];
        - head_handle: ControlPoint;
        - tail_handle: ControlPoint;
        - closed:      boolean.
        Note: "Butting" means that for any pair of successive arcs, the last
              control point of the former equals the first control point of
              the latter.
              In initialization no checks are done about this condition.
        """
        def __init__(self,
                     bezier_arcs=[],           # [BezierArc]
                     head_handle=None,         # ControlPoint
                     tail_handle=None,         # ControlPoint
                     closed=False,             # boolean
                     curve_name='BezierCurve', # string
                     ):
            self.curve_name = curve_name
            self.bezier_arcs = bezier_arcs
            if head_handle is None:
                self.head_handle = bezier_arcs[0].cp4[0]
            else:
                self.head_handle = head_handle
            if tail_handle is None:
                self.tail_handle = bezier_arcs[-1].cp4[-1]
            else:
                self.tail_handle = tail_handle
            self.closed = closed
        def __str__(self):
            s = self.curve_name + ': Bezier curve' + '; Bezier arcs:'
            count = 0
            for arc in self.bezier_arcs:
                s += '\n  arc '+ str(count)+ ': ' +str(arc)
                count += 1
            s += '\n  Head handle:    '+str(self.head_handle)
            s += '\n  Tail handle:    '+str(self.tail_handle)
            s += '\n  Closed:'    +str(self.closed)
            return s
        def bc2gs(self):
            """Conversion BezierCurve -> GimpStroke
            """
            cp_list = [self.head_handle]
            if len(self.bezier_arcs[0].cp4) > 1:
                for ba in self.bezier_arcs:
                    cp_list += ba.cp4[:3]
                cp_list.append(self.bezier_arcs[-1].cp4[-1])
            else:
                # Only one BezierArc and it has cp4 = [p0]
                # (may rise from Gimp stroke with only one anchor).
                cp_list.append(self.bezier_arcs[0].cp4[0])
            cp_list.append(self.tail_handle)
            return BCurve.GimpStroke(cp_list, self.closed, self.curve_name)
        def bc2vectors_object(self, image, name=None):
            """Conversion BezierCurve -> gimp.Vectors
            Makes vectors_object with 1 stroke.
            """
            if name is None:
                name = self.curve_name
            gs = self.bc2gs()             # BCurve.GimpStroke
            gv = BCurve.GimpVectors([gs]) # BCurve.GimpVectors
            return gv.gv2vectors_object(image, name)
    
    class BezierArc(object):
        """Data structure for one Bezier arc: arc determined by 4 control
           points.
        Attributes:
        - cp4:      [p0,p1,p2,p3] where each pi:ControlPoint, or
                    [p0] if only one control point (may rise from
                    a stroke with only one anchor);
        - arc_name: string
        """
        def __init__(self,
                     cp4,                 # cp4=[p0,p1,p2,p3] with pi:ControlPoint
                                          # or [p0]
                     arc_name='BezierArc' # string
                     ):
            self.cp4 = cp4
            self.arc_name = arc_name
        def reverse(self):
            self.cp4.reverse()
        def __str__(self):
            s = 'BezierArc, control points:'
            for cp in self.cp4:
                s += '\n  '+ str(cp)
            return s

# -----------------
# vectors_object2gv
# -----------------
# Conversion gimp.Vectors -> GimpVectors
# Args:
# - vectors_object: gimp.Vectors (Gimp vectors object)
# Returns:
# - GimpVectors
#
# Note: The inverse conversion is GimpVectors.gv2vectors_object(...).
#       See also BezierCurve.bc2vectors_object(...).
def vectors_object2gv(vectors_object):
    def xy_list2cp_list(xy_list): # xy_list = [x,y,x,y,...,]
        cp = []
        for i in range(0,len(xy_list),2):
            cp.append(BCurve.ControlPoint(xy_list[i], xy_list[i+1]))
        return cp
    stroke_list = []
    count = 0
    for stroke in vectors_object.strokes:
        xy_list = stroke.points[0]
        closed = stroke.points[1]
        cp_list = xy_list2cp_list(xy_list)
        stroke_name = vectors_object.name + ' stroke '+str(count)
        stroke_list.append(BCurve.GimpStroke(cp_list, closed, stroke_name))
        count += 1
    return BCurve.GimpVectors(stroke_list, vectors_object.name)

# ---------------------
# completed_bezier_arcs
# ---------------------
# Give bc:BCurve.BezierCurve, get the list of bezier arcs where, in the case of
# closed bc, a gap-closing arc is added at the end, with necessary modification
# for the case of closed 1-anchor stroke.
# Only copies of the original arcs are returned.
# Args:
# - bc:BCurve.BezierCurve
# Returns:
# - [BCurve.BezierArc]
def completed_bezier_arcs(bc):
    bas = deepcopy(bc.bezier_arcs)
    if not bc.closed:
        return bas
    gap_ba=BCurve.BezierArc(cp4=[bas[-1].cp4[-1],
                            bc.tail_handle,
                            bc.head_handle,
                            bas[0].cp4[0]])
    if len(bas[0].cp4) == 1: # 1-anchor stroke
        return [gap_ba]
    else:
        return bas + [gap_ba]

#==============================================================
#                Polynomial zeroes
#==============================================================

# A function to find zeroes of a polynomial, and as by-products the
# x-values of the extrema and of the inflection points.
# At the final phase the zeroes are found by binary search from monotonous
# segments.
# 
# Written just for fun. Other algorithms exist in abundance.
#
# In addition, as a simple wrapper, a function to do the same job when the
# polynomial is of degree 3 and given in the Bernstein basis.

# --------------------
# zeroes_of_polynomial
# --------------------
# Given a real polynomial, find the real zeroes, the x values of extrema,
# and the x values of the inflection points.
#
# Args: coefficients: list [c0,c1,...,cn] of the coefficients of the
#                     polynomial c0 + c1*x + ... + cn*(x**n);
#                     the ci are integer or float.
# Returns:
# 1. None in the case of a zero polynomial.
# 2. Otherwise:
#    - the list of real zeroes of the polynomial, in increasing order
#      (possibly []) with multiple roots repeated;
#    - the extrema (the x values);
#    - the inflection points (the x values).
def zeroes_of_polynomial(coefficients):
    from math import sqrt
    from functools import partial
    def f(x, coeff): # Horner
        s = 0.
        for ci in coeff[::-1]:
            s = ci + x*s
        return s
    coefficients = [float(ci) for ci in coefficients]
    # Find largest n with coefficients[n] non-zero and
    # discard all higher terms:
    n = -1
    for i in range(len(coefficients)):
        if coefficients[i] != 0.:
            n = i
    if n < 0: # Polynomial is identically zero.
        return None
    c = coefficients[:n+1] 
    # Make c[n] positive:
    if c[n] < 0:
        c = [-ci for ci in c]
    if n == 0: # Polynomial is a non-zero constant.
        return [],[],[]
    if n == 1: # Polynomial is c[0] + c[1]*x
        return [-c[0] / c[1]], [], []
    if n == 2: # Polynomial is c[0] + c[1]*x + c[2]*(x**2).
        discr = c[1]**2 - 4*c[0]*c[2]
        try:
            root = sqrt(discr)
            x1 = (-c[1] - root) / (2*c[2])
            x2 = (-c[1] + root) / (2*c[2])
            return [x1,x2], [-c[1]/(2*c[2])], []
        except ValueError:
            return [], [-c[1]/(2*c[2])], []
    # Build subdivision such that in each subinterval
    # the polynomial is monotonous:
    derivative = [e*i for e,i in enumerate(c)][1:]
    extrema, inflection, _ = zeroes_of_polynomial(derivative)
    x_far = sum([abs(ci/c[-1]) for ci in c])
    subdiv = [-x_far] + extrema + [x_far]
    # Solve and collect the zeroes of each monotonous segment:
    fc = partial(f, coeff=c)
    zeroes = []
    for i in range(len(subdiv)-1):
        lo, hi = subdiv[i], subdiv[i+1]
        x = zero_of_monotonous_function(fc, lo, hi,
                                    accuracy=1e-10, tolerance=1e-10)
        if not(x is None):
            zeroes.append(x)
    return zeroes, extrema, inflection

# ------------------------------
# zeroes_of_bernstein_polynomial
# ------------------------------
# Given a polynomial of degree 3, in Bernstein basis,
# find the real zeroes, the x values of extrema, and the x values of the
# inflection points.
#
# Args: coefficients: list [c0,c1,c2,c3] of the coefficients of the
#                     polynomial c0*b0(t) + c1*b1(t) + c2*b2(t) + c3*b3(t)
#                     where b0,b1,b2,b3 are the bernstein polynomial of degree 3,
#                     and the ci are integer or float.
#                     Must be 4 coefficient.
# Returns:
# 1. None in the case of a zero polynomial.
# 2. Otherwise:
#    - the list of real zeroes of the polynomial, in increasing order
#      (possibly []) with multiple roots repeated;
#    - the extrema (the x values);
#    - the inflection points (the x values).
def zeroes_of_bernstein_polynomial(coefficients):
    # Convert a polynomial f(t) from the Bernstein basis {b0,b1,b2,b3}
    # to the power basis {1,t,t**2,t**3}.
    # Args:
    # - coefficients: list [c0,c1,c2,c3] = coefficients of a polynomial in
    #   Bernstein basis: so, the polynomial is
    #   f(t) = c0*b0(t) + c1*b1(t) + c2*b2(t) + c3*b3(t).
    # Returns:
    # - a list [a0,a1,a2,a3]: coefficients of f(t) in power basis:
    #   f(t) = a0 + a1*t + a2*(t**2) + a3*(t**3).
    def bernstein2power(coefficients):
        c0,c1,c2,c3 = [float(c) for c in coefficients]
        cb0 = [c0*b for b in [1,-3, 3,-1]]
        cb1 = [c1*b for b in [0, 3,-6, 3]]
        cb2 = [c2*b for b in [0, 0, 3,-3]]
        cb3 = [c3*b for b in [0, 0, 0, 1]]
        result = []
        for i in range(4):
            result.append(cb0[i] + cb1[i] + cb2[i] + cb3[i])
        return result
    if len(coefficients) != 4:
        s = "zeroes_of_bernstein_polynomial: "
        s += "wrong number of coefficients"
        raise Exception(s)
    power_basis_coefficients = bernstein2power(coefficients)
    return zeroes_of_polynomial(power_basis_coefficients)

# ---------------------------
# zero_of_monotonous_function
# ---------------------------
# Find the zero, if any, of a monotonous function f(x)
# in the interval [lo,hi]. If none is found, None is returned.
# Args:
# - f, lo, hi: the monotonous function with start and end value;
# - accuracy:  accuracy for x.
# - tolerance: float>=0; if no zeroes exist in the interval, then
#              if f(lo) or f(hi) is less than 'tolerance' from 0,
#              then lo or hi is returned, respectively.
# Returns: the zero or None.
def zero_of_monotonous_function(f, lo, hi, accuracy=1e-10, tolerance=1e-10):
    MAX_ROUNDS = 100
    lo, hi = float(lo), float(hi)
    flo, fhi = float(f(lo)), float(f(hi))
    if (flo > 0.) == (fhi > 0.):
        if abs(flo) <= tolerance:
            return lo
        elif abs(fhi) <= tolerance:
            return hi
        else:
            return None
    if flo == 0.:
        return lo
    if fhi == 0.:
        return hi
    count = 0
    while hi-lo > accuracy:
        mid = (lo+hi)/2.
        fmid = float(f(mid))
        if fmid == 0.:
            return mid
        if (flo > 0.) == (fmid > 0.):
            lo, flo = mid, fmid
        else:
            hi, fhi = mid, fmid
        count += 1
        if count > MAX_ROUNDS:
            break
    x = (fhi*lo - flo*hi) / (fhi-flo)
    return x

# -------------------------
# solve_bernstein2_equation
# -------------------------
# Find real zeroes of the equation c0*b0(t) + c1*b1(t) + c2*b2(t) = 0 where
# b0,b1,b2 are the Bernstein polynomials of degree 2 and
# c0,c1,c2 are real coefficients.
# Return the zeroes as a list (0, 1, or 2 items).
# Exception: If the function is constant 0, return None.
# Args:
# - c0,c1,c2: float
# Returns:
# - None or list of floats (possibly empty)
def solve_bernstein2_equation(c0,c1,c2):
    from math import sqrt
    ZERO = 1e-14
    if abs(c0 - 2.*c1 + c2) < ZERO: # equation of degree <= 1
        try:
            return [c0 / (2.*(c0 - c1))]
        except ZeroDivisionError: # equation of degree 0
            if abs(c0) < ZERO: # constant 0
                return None
            else: # constant not 0
                return []
    try:
        root = sqrt(c1**2 - c0*c2)
    except ValueError: # No real roots
        return []
    return [(c0 - c1 + root) / (c0 - 2.*c1 + c2),
            (c0 - c1 - root) / (c0 - 2.*c1 + c2)]

# --------------------------
# point_on_linear_Bezier_arc
# --------------------------
# Assume that a Bezier arc B(t) is on a straight line, control points p0,p1,p2,p3.
# solve the pairs (t,B(t)) where B(t) = (1-q)*p0 + q*p3 (so that B(t)-p0 = q*(p3-p0)).
# Args:
# - cp4: [complex,complex,complex,complex] (=[p0,p1,p2,p3])
# - q:   float
# Returns:
# [(float,complex)] (= [(t,B(t))])
def point_on_linear_Bezier_arc(cp4, q):
    p0,p1,p2,p3 = cp4
    p01 = -p0+p1
    p02 = -p0+p2
    p03 = -p0+p3
    c0 = -q*vdot(p03,p03)
    c1 = vdot(p01-q*p03, p03)
    c2 = vdot(p02-q*p03, p03)
    c3 = vdot(p03-q*p03, p03)
    polynomial_zeroes = zeroes_of_bernstein_polynomial([c0,c1,c2,c3])
    if polynomial_zeroes is None: # Polynomial identically 0
        raise Exception("A zero-length arc?")
    zeroes, extrema, inflection = polynomial_zeroes
    return [(t, bezier_rv(t,cp4)) for t in zeroes]

#==============================================================
#                         Draw
#==============================================================
# ------------------------
# gimp_draw_vectors_object
# ------------------------
# The Gimp version of the routine 'drawing_routine_for_bezier':
# Insert into Gimp the input vectors object, and make it visible if
# visible=True.
# Args:
# - image: Gimp's image (obtained usually from GUI);
# - vectors_object: a vectors object in Gimp;
# - visible: Boolean.
# Returns the vectors object in Gimp, and makes it visible if visible=True.
def gimp_draw_vectors_object(image, vectors_object, visible=True):
    # Insert a Gimp vectors object in Gimp.
    # Adapted from http://registry.gimp.org/files/lqr_wpset-0.14.py.txt
    # by Mike Kazantsev.
    def gimp_insert_vectors(image, vectors_object, position=0):
        if gimp.version >= (2, 8, 0):
            # 2.7.0 has gimp-image-insert-vectors, but fails with
            # parent=None
            parent = None
            return pdb.gimp_image_insert_vectors(
                     image, vectors_object, parent, position)
        else:
            return pdb.gimp_image_add_vectors(
                     image, vectors_object, position)# Deprecated!
    gimp_insert_vectors(image, vectors_object)
    vectors_object.visible = visible
    return vectors_object

# For debugging only.
# Draw a Bezier arc from control points p0,p1,p2,p3.
# - cp4: [complex,complex,complex,complex]
def gimp_draw_bezier_arc_cp4(image, cp4, name='Bezier arc'):
    vcp4 = [[z.real,z.imag] for z in cp4]
    cp6 = [vcp4[0]] + vcp4 + [vcp4[-1]]
    xy = cp6[0] + cp6[1] + cp6[2] + cp6[3] + cp6[4] + cp6[5]
    pdb.gimp_image_undo_group_start(image)
    vectors = pdb.gimp_vectors_new(image, name)
    pdb.gimp_vectors_stroke_new_from_points(vectors, 0, len(xy), xy, False)
    pdb.gimp_image_insert_vectors(image, vectors, None, 0)
    vectors.visible = True
    pdb.gimp_image_undo_group_end(image)

# Gimp error message function.
# The input 'message' is a string.
def gimp_error_message(message, print_message=False):
    if print_message:
        print(message)
    # 0 = message box; 1 = console; 2 = error console
    pdb.gimp_message_set_handler(2) # Laita 0 lopulliseen!
    pdb.gimp_message(message)

#==============================================================
#                     Plane vectors
#==============================================================

# Make similitude where z0,z1 are mapped onto w0,w1.
def make_similitude(z0,z1,w0,w1):
    def similitude(z):
        return (w1*(z-z0) - w0*(z-z1)) / (-z0+z1)
    return similitude

def vdot(v,w):
    return v.real*w.real + v.imag*w.imag

# v dot (w perpendicular)
def cross(v,w):
    return (v*w.conjugate()).imag

def is_straight_line(cp4):
    ZERO = 1e-6
    p0,p1,p2,p3 = cp4
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    p30 = -p3+p0 # chord
    return ((cross(p01,p30)<ZERO) and (cross(p12,p30)<ZERO) and (cross(p23,p30)<ZERO))

#==============================================================
#                        Bezier
#==============================================================
# ---------
# bezier_rv
# ---------
# The usual Bezier curve from control points.
# Version where all points are complex numbers
# (the plane is the complex number plane).
# Args:
# - t:  float
# - cp4:[complex,complex,complex,complex]
# Returns:
# - complex
def bezier_rv(t, cp4):
    P0,P1,P2,P3 = cp4
    if t > 0.5:
        u = (1-t)/t
        t3 = t**3
        return t3*(u*(u*(u*P0+3*P1)+3*P2)+P3) # complex
    else:
        u = t/(1-t)
        t3 = (1-t)**3
        return t3*(u*(u*(u*P3+3*P2)+3*P1)+P0) # complex

# Derivative: Returns complex.
def bezier_dot_rv(t, control_points):
    P0,P1,P2,P3 = control_points
    P01 = -P0+P1
    P12 = -P1+P2
    P23 = -P2+P3
    if t > 0.5:
        u = (1-t)/t
        t2 = t**2
        return 3*t2*(u*(u*P01+2*P12)+P23) # complex
    else:
        u = t/(1-t)
        t2 = (1-t)**2
        return 3*t2*(u*(u*P23+2*P12)+P01) # complex

# Second derivative
# Args:
# - t:  float
# - cp4:[complex,complex,complex,complex]
# Returns:
# - complex
def bezier_dot_dot_rv(t, control_points): # Could be made faster as above.
    p0,p1,p2,p3 = control_points
    return -6*(1-t)*(-p0+p1) + 6*(1-2*t)*(-p1+p2) + 6*t*(-p2+p3)

# -------------------------
# bezier_new_control_points
# -------------------------
# Given a Bezier arc B(t) and two distinct parameter values a and b,
# find a new parametrization C(t) of the same curve such that
# B(a) = C(0) and B(b) = C(1).
# Args:
# - a: float;
# - b: float;
# - control_points: the list [p0,p1,p2,p3] of control points (pi:complex).
# Returns:
# - new control points [q0,q1,q2,q3]:[RowVector] such that the corresponding
#   Bezier curve C(t) is precisely the same as B as a curve
#   but B(a) = C(0) and B(b) = C(1).
# Notes:
# - This enables one to extend or to shorten the given Bezier arc.
# - There is an exception: If a=b, the returned control points
#   are equal, so the new curve is a single point, not the original
#   curve. The function gives no signal about such occurence.
def bezier_new_control_points(a, b, control_points):
    def bezier_2_rv(t, r0, r1, r2): # Bezier curve of degree two (complex)
        b20 = (1-t)**2
        b21 = 2*(1-t)*t
        b22 = t**2
        return b20*r0 + b21*r1 + b22*r2
    p0,p1,p2,p3 = control_points
    q0 = bezier_rv(a, control_points)
    q1 = (1-b)*bezier_2_rv(a, p0,p1,p2) + b*bezier_2_rv(a, p1,p2,p3)
    q2 = (1-a)*bezier_2_rv(b, p0,p1,p2) + a*bezier_2_rv(b, p1,p2,p3)
    q3 = bezier_rv(b, control_points)
    return [q0,q1,q2,q3]

# ------------------------------
# bezier_subdivide_at_parameters
# ------------------------------
# Given a Bezier arc B(t) and a list of parameter values t1,t2,...,
# find control points for all subarcs when B(t), 0<=t<=1, is subdivided
# at parameter values t1,...
# The input Bezier arc is given as its 4 control points [p0,p1,p2,p3],
# and the output is a list of new sets of 4 control points [[q0,q1,q2,q3],...].
# The plane is viewed as the complex number plane.
# Args:
# - parameter_list: [float]                           ([t1,t2,...])
# - control_points: [complex,complex,complex,complex] ([p0,p1,p2,p3])
# Returns:
# - [[complex,complex,complex,complex]]               ([[q0,q1,q2,q3],...])
def bezier_subdivide_at_parameters(parameter_list, control_points):
    ts01 = [t for t in parameter_list if 0 < t < 1] # Drop t's outside (0,1)
    ts01.sort()
    try:
        ts = [ts01[0]]
    except IndexError: # Empty parameter list
        return [[p for p in control_points]] # copy
    for t in ts01[1:]: # Drop duplicates.
        if t != ts[-1]:
            ts.append(t)
    ts = [0]+ts+[1] # Add t=0,1
    # Collect pairs [t(i),t(i+1)]
    abs = []
    for i in range(1,len(ts)):
        abs.append([ts[i-1],ts[i]])
    arc_cps = []
    for a,b in abs:
        arc_cps.append(bezier_new_control_points(a, b, control_points))
    return arc_cps

# -------------------------
# bezier_arc_end_directions
# -------------------------
# Given a Bezier arc (control points, 0<=t<=1), find, at each end,
# the unit direction vector of the arc.
# Note: this does not mean the derivatives of the parametric representation,
# but non-zero vectors which are tangential to the arc and point in the
# direction where the arc is located when viewed from that end point.
# Return the result as two complex numbers (= two vectors),
# or as None if the arc is one point.
# The plane is viewed as the complex number plane, so points are complex numbers.
# Args:
# - cp4: [complex,complex,complex,complex] (control points)
# Returns: either None (when the arc is one point), or
# - complex (unit direction vector at starting point)
# - complex (unit direction vector at ending point)
# Note: If len(cp4)=1, None is returned. That case comes from a 1-anchor stroke.
def bezier_arc_end_directions(cp4):
    ZERO = 1e-12
    if len(cp4) == 1: # May happen in a 1-anchor stroke
        return None
    try:
        p0,p1,p2,p3 = cp4
    except ValueError:
        raise Exception("bezier_arc_end_directions: Should not happen 1!")
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    if vdot(p01,p01) > ZERO:
        head_dir = p01
    elif vdot(p12,p12) > ZERO:
        head_dir = p12
    elif vdot(p23,p23) > ZERO:
        head_dir = p23
    else: # Arc is one point
        return None
    if vdot(p23,p23) > ZERO:
        tail_dir = -p23
    elif vdot(p12,p12) > ZERO:
        tail_dir = -p12
    elif vdot(p01,p01) > ZERO:
        tail_dir = -p01
    else:
        raise Exception("bezier_arc_end_directions: Should never happen 2!")
    return head_dir/abs(head_dir), tail_dir/abs(tail_dir)

# ---------------------
# bezier_extremal_points
# ---------------------
# Given a Bezier curve (4 control points p0,p1,p2,p3:complex), find its
# extremal points.
# Return list [[t,B(t)]] where B(t) is an extremal point and t is the
# parameter value.
# By an extremal point we understand roughly a point on the curve which is
# locally most distant from the chord. The exact condition is B'(t)||p0p3.
# Args:
# - cp4: [complex,complex,complex,complex]
# Returns:
# - [[float,complex]]
def bezier_extremal_points(cp4, restrict01=True):
    p0,p1,p2,p3 = cp4
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    p30 = -p3+p0 # chord
    p01xp03 = cross(p01,p30)
    p12xp03 = cross(p12,p30)
    p23xp03 = cross(p23,p30)
    ts = solve_bernstein2_equation(p01xp03, p12xp03, p23xp03)
    if ts is None:
        return []
    if restrict01:
        ts = [t for t in ts if 0<=t<=1]
    return [[t,bezier_rv(t,cp4)] for t in ts]

# ---------------------
# bezier_side_extremals
# ---------------------
# Extension of 'bezier_extremal_points' except that assumes that
# - there are no inflection points (direction of curvature does not change);
# - there is a new argument 'level' (default=0).
# If level=0 works just as 'bezier_extremal_points' except that instead of
# a list returns now the extremal point [(t,B(t))] (1-item list).
# It is unique since there are no inflection points.
# If level=1 returns [(t0,B(t0)), (t,B(t)), (t1,B(t1))] where
# - 0 < t0 < t < t1 < 1;
# - (t0,B(t0))   is the extremal point for the arc from B(0) to B(t);
# - (t,B(t)) is the extremal point;
# - (t1,B(t1)) is the extremal point for the arc from B(t) to B(1).
# For a straight line segment, pairs (t,B(t)) are returned where B(t) is
# at distances 0.5, 0.25, 0.75 of the line segment
# Currently this is how far it goes.
# In the future, maybe the scheme is extended recursively, giving a nice
# subdivision for the whole Bezier arc cp4.
# Failures:
# - In case of failure to find (t,B(t)) raises Exception.
# - In case of failure to find (t0,B(t0)), None is returned in place of (t0,B(t0)).
# - In case of failure to find (t1,B(t1)), None is returned in place of (t1,B(t1)).
# 
# Args:
# - cp4:[complex,complex,complex,complex]
# Returns:
# - [None or (float,complex)] ( = [None or (t,B(t))] )
def bezier_side_extremals(cp4, level=0):
    p0,p1,p2,p3 = cp4
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    p30 = -p3+p0 # chord
    if is_straight_line(cp4):
        tBs_05  = [tB for tB in point_on_linear_Bezier_arc(cp4, 0.5)  if 0<tB[0]<1]
        tBs_025 = [tB for tB in point_on_linear_Bezier_arc(cp4, 0.25) if 0<tB[0]<1]
        tBs_075 = [tB for tB in point_on_linear_Bezier_arc(cp4, 0.75) if 0<tB[0]<1]
        if (len(tBs_05) != 1) or (len(tBs_025) != 1) or (len(tBs_075) != 1):
            raise Exception("Should not happen!")
        return [tBs_025[0], tBs_05[0], tBs_075[0]]
    ts = solve_bernstein2_equation(cross(p01,p30), cross(p12,p30), cross(p23,p30))
    if ts is None:
        raise Exception("bezier_side_extremals: len(ts) = 0. Straight line though is_straight_line returns False?")
    ts = [t for t in ts if 0<t<1]
    if len(ts) != 1:
        raise Exception("bezier_side_extremals: len(ts) = "+str(len(ts))+". Inflection point?")
    t_ext = ts[0]
    B_ext = bezier_rv(t_ext,cp4)
    if level == 0:
        return [(t_ext,B_ext)]
    elif level == 1:
        p0Bext = -p0 + B_ext
        ts = solve_bernstein2_equation(cross(p01,p0Bext), cross(p12,p0Bext), cross(p23,p0Bext))
        if ts is None:
            t0B0 = None
        ts = [t for t in ts if 0<t<1]
        if len(ts) != 1:
            t0B0 = None
        else:
            t0 = ts[0]
            B0 = bezier_rv(t0,cp4)
            t0B0 = (t0,B0)
        Bextp3 = -B_ext + p3
        ts = solve_bernstein2_equation(cross(p01,Bextp3), cross(p12,Bextp3), cross(p23,Bextp3))
        if ts is None:
            t1B1 = None
        ts = [t for t in ts if 0<t<1]
        if len(ts) != 1:
            t1B1 = None
        else:
            t1 = ts[0]
            B1 = bezier_rv(t1,cp4)
            t1B1 = (t1,B1)
        return [t0B0, (t_ext,B_ext), t1B1]
    else:
        raise Exception("bezier_side_extremals: level > 1 not implemented yet.")

# --------------------
# bezier_closest_point
# --------------------
# Find closest point (parameter value t and the point B(t)) to a given v on
# the Bezier curve B(t) with control points cp4.
# If restrict01=True, accept only values 0<=t<=1 and include the end points
# t=0,1 among the tested values.
# Args:
# - cp4: [complex,complex,complex,complex] (control points)
# - v:   complex
# - restrict01: boolean
# Returns:
# - [float,complex] (=[t,B(t)])
def bezier_closest_point(cp4, v, restrict01):
    t_Bs = bezier_closest_point_candidates(cp4, v)
    if restrict01:
        t_Bs = [[t,B] for t,B in t_Bs if 0<t<1]
        t_Bs += [[0,cp4[0]], [1,cp4[-1]]]
    return min(t_Bs, key=(lambda x: abs(x[1]-v)))

# -------------------------------
# bezier_closest_point_candidates
# -------------------------------
# Candidates for closest points: Solutions to (B(t)-v).B'(t) = 0.
# Args:
# - cp4: [complex,complex,complex,complex] (control points)
# - v:   complex
# Returns:
# - [[float,complex]] (=[[t,B(t)]])
def bezier_closest_point_candidates(cp4, v):
    p0,p1,p2,p3 = cp4
    # Control points w0,w1,w2,w3 when the basis of 1, t, t^2/2, t^3/6
    # is used instead of the Bernstein polynomials:
    w0 = p0
    w1 = 3*(-p0 + p1)
    w2 = 6*(p0 - 2*p1 + p2)
    w3 = 6*(-p0 + 3*p1 - 3*p2 + p3)
    # Dot products:
    w0w1 = w0.real * w1.real + w0.imag * w1.imag
    w0w2 = w0.real * w2.real + w0.imag * w2.imag
    w0w3 = w0.real * w3.real + w0.imag * w3.imag
    w1w1 = w1.real * w1.real + w1.imag * w1.imag
    w1w2 = w1.real * w2.real + w1.imag * w2.imag
    w1w3 = w1.real * w3.real + w1.imag * w3.imag
    w2w2 = w2.real * w2.real + w2.imag * w2.imag
    w2w3 = w2.real * w3.real + w2.imag * w3.imag
    w3w3 = w3.real * w3.real + w3.imag * w3.imag
    #
    vw1 = v.real * w1.real + v.imag * w1.imag
    vw2 = v.real * w2.real + v.imag * w2.imag
    vw3 = v.real * w3.real + v.imag * w3.imag
    # Polynomial coefficients:
    c0 = w0w1 - vw1
    c1 = w0w2 + w1w1 - vw2
    c2 = w0w3/2 + 3*w1w2/2 - vw3/2
    c3 = 2*w1w3/3 + w2w2/2
    c4 = 5*w2w3/12
    c5 = w3w3/12
    # Solve polynomial zeroes:
    zeroes_etc = zeroes_of_polynomial([c0,c1,c2,c3,c4,c5])
    if zeroes_etc is None:
        m = "bezier_closest_point_candidates: No polynomial zeroes for:"
        m += "\nv   = "+str(v)
        m += "\ncp4 = "+str(cp4)
        raise Exception(m)
    ts = zeroes_etc[0]
    #print("ts = "+str(ts))
    # Compute B(t)'s:
    t_Bs = [[t, w0 + t*w1 + t*t*w2/2 + t*t*t*w3/6] for t in ts]
    return t_Bs

# ---------------------
# bezier_special_points
# ---------------------
# Find the special points of a cubic Bezier curve:
# inflection points, cusp points, self-intersection points.
#
# Args:
# - control_points: [p0,p1,p2,p3]
#
# Returns: (case, param_list), which is one of:
# (-1, [t1,t2]) when there is a loop or a cusp point, with
#               t1,t2 the parameter values with B(t1)=B(t2)
#               (ideally, t1=t2 in the case of a cusp);
# (0, [])       when the curve is degenerate: a parabola or
#               contained on a straight line;
# (1, [t])      when there is a single inflection point, with
#               t the parameter of the inflection point.
# (2, [t1,t2])  when there are two inflection points, with
#               t1,t2 the parameters for the inflection points.
#
# Note: The algorithm works as if infinite computation precision were
#   possible. The caller had better check the following:
# - In the case of a loop or of two inflection points (cases -1 or 2):
#   If the returned parameter values t1,t2 are very close, it may be
#   a sign of a cusp point, instead. (Or maybe the caller should take
#   it as a cusp point.)
# - In the case of inflection points (cases 1 or 2):
#   If one or both of the returned parameter values has very large
#   absolute value, it may mean that the inflection point is at infinity
#   (so, non-existent).
def bezier_special_points(control_points):
    from math import sqrt
    p0,p1,p2,p3 = control_points
    w1 = -p0 - p1 + p2 + p3
    w2 = p0 - p1 - p2 + p3
    w3 = -p0 + 3*p1 - 3*p2 + p3
    C1 = 18.0 * (-w2.real*w3.imag + w2.imag*w3.real)
    C2 = 4.50 * (-w3.real*w1.imag + w3.imag*w1.real)
    C3 = 2.25 * (-w1.real*w2.imag + w1.imag*w2.real)
    try:
        X = C2/C1
        Y = C3/C1
        Z = -X**2 + 2*Y
        try:
            root = sqrt(3*Z)
            t1 = .5 + X + root
            t2 = .5 + X - root
            return -1, [t1,t2] # loop or cusp
        except ValueError: # Z < 0
            root = sqrt(-Z)
            t1 = .5 + X + root
            t2 = .5 + X - root
            return 2, [t1,t2] # two inflection points
    except ZeroDivisionError: # C1 = 0
        try:
            t = .5 + C3/C2
            return 1, [t] # one inflection point
        except ZeroDivisionError: # C1 = C2 = 0
            return 0,[] # parabola or everything on straight line.

# ----------------------------
# bezier_unit_direction_vector
# ----------------------------
# Find the unit direction vector of a Bezier curve B(t) at parameter value t.
# The direction is where the curve is going when t grows.
# For a cusp this means the direction after the cusp!
# Return also the case:
# - case = 0:  the curve is constant (and the returned vector is None);
# - case = 1:  regular point: the vector was computed from B'(t);
# - case = 2:  cusp: the vector was computed from B''(t);
# - case = 3:  B'(t)=B''(t)=0: the vector was computed from B'''(t).
# Args:
# - t:              float
# - control_points: [complex,complex,complex,complex]
# Returns:
# - complex or None (None is returned when B(t) is constant).
# - int             (case)
def bezier_unit_direction_vector(t, control_points):
    p0,p1,p2,p3 = control_points
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    b2 = [(1-t)**2, 2*(1-t)*t, t**2]              # Bernstein 2
    dot = b2[0]*p01 + b2[1]*p12 + b2[2]*p23       # B'(t)/3
    try:
        return dot/abs(dot), 1                    # case = 1
    except ZeroDivisionError:
        dotdot = -(1-t)*p01 + (1-2*t)*p12 + t*p23 # B''(t)/6
        try:
            return dotdot/abs(dotdot), 2          # case = 2
        except ZeroDivisionError:
            dotdotdot = p01 - 2*p12 + p23         # B'''(t)/6
            try:
                return dotdotdot/abs(dotdotdot), 3 # case = 3
            except ZeroDivisionError:
                return None, 0                    # case = 0

# --------------------------------
# extend_path_stroke_with_straight
# --------------------------------
# Given a path and its open stroke, extend the stroke at each end
# by the given amount, measured along the stroke,
# continuing the curve with a straight line segment.
# The input extension lengths can be either
# - absolute, or
# - relative to the length of the stroke.
# A negative input extension results in no extension: the stroke
# is never shortened at either end.
# Return a new path with the extended stroke as its sole stroke.
# The new stroke has the same anchors as the original and additional anchors
# at the ends of the straight extensions (possibly the same).
# Return also the absolute lengths of the extensions.
# Head and tail handles are not preserved (set to 0) even if no extension was done.
# Args:
# - image
# - path:                gimp.Vectors
# - stroke:              gimp.VectorsBezierStroke (a stroke of the path)
# - left_extension:      float          (required amount of left extension)
# - right_extension:     float          (required amount of right extension)
# - extensions_absolute: boolean (if True, left_extension and right_extension
#                                 are taken to be absolute (pixels),
#                                 otherwise relative)
# - accuracy:            float   (very inaccurate way to control the excess in the
#                                 lengths of the constructed extensions)
# Returns:
# - gimp.Vectors     (1-stroke path)
# - float            (absolute extension length on the left)
# - float            (absolute extension length on the right)
def extend_path_stroke_with_straight(image,
                                     path,
                                     stroke,
                                     left_extension,
                                     right_extension,
                                     extensions_absolute=True,
                                     ):
    PRECISION = 1e-10
    if stroke.points[1]:
        raise Exception("extend_path_stroke_with_straight: Expected an open stroke, got closed")
    ori_gv = vectors_object2gv(path)
    ori_length = pdb.gimp_vectors_stroke_get_length(path, stroke.ID, PRECISION)
    target_left_ext = max(0, left_extension)
    target_right_ext = max(0, right_extension)
    if not extensions_absolute: # relative: make the lengths absolute:
        target_left_ext *= ori_length
        target_right_ext *= ori_length
    for index, gstroke in enumerate(path.strokes):
        if gstroke.ID == stroke.ID:
            ori_gs = ori_gv.stroke_list[index] # BCurve.GimpStroke
            break
    else:
        raise Exception("extend_path_stroke_with_straight: Not found the stroke!")
    ori_bc = ori_gs.gs2bc()                # BCurve.BezierCurve
    ori_bas = ori_bc.bezier_arcs           # [BCurve.BezierArc]
    if len(ori_bas) == 0:
        raise Exception("extend_path_stroke_with_straight: No segments in the stroke!")
    first_arc, last_arc = ori_bas[0], ori_bas[-1] # Same if only_one_arc=True
    # The following unit vectors point from the end point towards the curve:
    start_direction,_ = bezier_arc_end_directions(first_arc.cp4)
    _,end_direction = bezier_arc_end_directions(last_arc.cp4)
    P,Q = first_arc.cp4[0], last_arc.cp4[-1] # First and last anchor
    new_first_cp4 = [P - target_left_ext * start_direction,
                     P - 2 * target_left_ext * start_direction / 3,
                     P - target_left_ext * start_direction / 3,
                     P]
    new_last_cp4 = [Q,
                    Q - target_right_ext * end_direction / 3,
                    Q - 2 * target_right_ext * end_direction / 3,
                    Q - target_right_ext * end_direction]
    #if target_left_ext == 0:
    if target_left_ext < PRECISION:
        new_left_bas = []
    else:
        new_left_bas = [BCurve.BezierArc(cp4 = new_first_cp4)]
    #if target_right_ext == 0:
    if target_right_ext < PRECISION:
        new_right_bas = []
    else:
        new_right_bas = [BCurve.BezierArc(cp4 = new_last_cp4)]
    new_bas = new_left_bas + ori_bas + new_right_bas
    new_bc = BCurve.BezierCurve(bezier_arcs = new_bas)
    new_path = new_bc.bc2vectors_object(image, name='extended')
    return new_path, target_left_ext, target_right_ext

#======================================================
#                   Bounding box
#======================================================

# -----
# gv_BB
# -----
# Bounding box of Gimp's path, given as an object:BCurve.GimpVectors.
# Return as list [NW, SE] of two corner points,
# where NW is top left, SE is bottom right. (The directions are indeed as
# on the screen: N is up, S is down, provided that the argument 'Gimp' is correct!)
# Args:
# - gv: BCurve.GimpVectors
# - Gimp: boolean
# Returns:
# - [NW, SE]: [complex,complex]
def gv_BB(gv, Gimp=True):
    hor = gv_tangent_points(gv, 1+0j)
    ver = gv_tangent_points(gv, 0+1j)
    anchors = gv_anchors(gv, only_first_stroke=False)
    xmax = max([v.real for v in ver+anchors])
    xmin = min([v.real for v in ver+anchors])
    ymax = max([v.imag for v in hor+anchors])
    ymin = min([v.imag for v in hor+anchors])
    if Gimp:
        return [complex(xmin,ymin), complex(xmax,ymax)]
    else:
        return [complex(xmin,ymax), complex(xmax,ymin)]

# -----------------
# gv_tangent_points
# -----------------
# Given Gimp's path as an object:BCurve.GimpVectors
# and a non-zero direction vector d,
# find the points where some arc is parallel to d.
# Include cusps.
# Exception: For arcs which are all the way parallel, the point t=0.5 is chosen.
# Args:
# - gv: BCurve.GimpVectors
# - direction: complex (assumed non-zero)
# Returns:
# - [complex] (touching points)
def gv_tangent_points(gv, direction):
    #gv = vectors_object2gv(path_vectors) # BCurve.GimpVectors
    points = []
    for gs in gv.stroke_list:  # BCurve.GimpStroke
        bc = gs.gs2bc()         # BCurve.BezierCurve
        points += BezierCurve_tangent_points(bc, direction)
    return points

# -----------------
# gv_tangent_points
# -----------------
# Given Gimp's path as an object:BCurve.GimpVectors, find the anchors
# Include cusps.
# Exception: For arcs which are all the way parallel, the point t=0.5 is chosen.
# Args:
# - gv: BCurve.GimpVectors
# - only_first_stroke: boolean
# Returns:
# - [complex]
def gv_anchors(gv, only_first_stroke=False):
    anchors = []
    if only_first_stroke:
        strokes = gv.stroke_list[0]
    else:
        strokes = gv.stroke_list
    for gs in strokes:
        for i,cp in enumerate(gs.cp_list):
            if i%3 == 1:
                anchors.append(cp)
    return anchors

# --------------------------
# BezierCurve_tangent_points
# --------------------------
# Given a Bezier curve:Bcurve.BezierCurve (essentially list of butting
# Bezier arcs) and a non-zero direction vector d,
# find the points on B where the curve is parallel to d.
# Includes cusps and isolated anchors.
# For arcs which are all the way parallel to d, the point
# with t=0.5 is returned
# Args:
# - bcurve: BCurve.BezierCurve
# - direction: complex (assumed non-zero)
# Returns:
# - [complex]
def BezierCurve_tangent_points(bcurve, direction):
    if not bcurve.closed:
        balist = bcurve.bezier_arcs
    else: # closed
        # In Gimp a closed stroke is only marked to be closed.
        # We must append one more arc to close the gap:
        gap_ba = BCurve.BezierArc(cp4 = [bcurve.bezier_arcs[-1].cp4[-1],
                                         bcurve.tail_handle,
                                         bcurve.head_handle,
                                         bcurve.bezier_arcs[0].cp4[0]])
        balist = bcurve.bezier_arcs + [gap_ba]
    cp4s = [ba.cp4 for ba in balist]
    points = []
    for cp4 in cp4s:
        if len(cp4) == 1: # A one-anchore stroke
            points.append(cp4[0])
            continue
        tBs = bezier_curve_tangent_points(cp4, direction, restrict01=True)
        if tBs is None: # arc is straight line segment parallel to direction
            t = 0.5
            p = bezier_rv(t, cp4)
            points.append(p)
            #t = 1.
            #p = bezier_rv(t, cp4)
            #points.append(p)
        else:
            for t,p in tBs:
                points.append(p)
    return points

# ---------------------------
# bezier_curve_tangent_points
# ---------------------------
# Given an infinite Bezier curve B(t) (4 control points)
# and a non-zero direction vector d,
# find the points on B with B'(t) parallel to d.
# If restrict01=True, accept only those with 0<=t<=1.
# Return as a list of pairs [t,p] where t is the parameter value (float)
# and p is the touching point B(t) (complex).
# Exception: If d=0, or if the curve is contained on a line parallel to d,
# or if len)control_points) is not 4 (1-anchor stroke!),
# return None.
# Args:
# - control_points: [complex,complex,complex,complex]
# - direction:      complex (assumed non-zero)
# - restrict01:     boolean
# Returns: None or
# - list [[t,p]]: [[float,complex]]   (=[[t,B(t)],...])
def bezier_curve_tangent_points(control_points, direction, restrict01=True):
    if len(control_points) < 4:
        return None
    p0,p1,p2,p3 = control_points
    p01 = -p0+p1
    p12 = -p1+p2
    p23 = -p2+p3
    # pij dot (direction perpendicular):
    p01d = -p01.real * direction.imag + p01.imag * direction.real
    p12d = -p12.real * direction.imag + p12.imag * direction.real
    p23d = -p23.real * direction.imag + p23.imag * direction.real
    #
    ts = solve_bernstein2_equation(p01d,p12d,p23d)
    if ts is None:
        return None
    if restrict01:
        ts = [t for t in ts if 0<=t<=1]
    tBs = [[t, bezier_rv(t, control_points)] for t in ts]
    return tBs

# -------------------------
# solve_bernstein2_equation
# -------------------------
# Find real zeroes of the equation c0*b0(t) + c1*b1(t) + c2*b2(t) = 0 where
# b0,b1,b2 are the Bernstein polynomials of degree 2 and
# c0,c1,c2 are real coefficients.
# Return the zeroes as a list (0, 1, or 2 items).
# Exception: If the function is constant 0, return None.
# Args:
# - c0,c1,c2: float
# Returns:
# - None or list of floats (possibly empty)
def solve_bernstein2_equation(c0,c1,c2):
    from math import sqrt
    ZERO = 1e-14
    if abs(c0 - 2.*c1 + c2) < ZERO: # equation of degree <= 1
        try:
            return [c0 / (2.*(c0 - c1))]
        except ZeroDivisionError: # equation of degree 0
            if abs(c0) < ZERO: # constant 0
                return None
            else: # constant not 0
                return []
    try:
        root = sqrt(c1**2 - c0*c2)
    except ValueError: # No real roots
        return []
    return [(c0 - c1 + root) / (c0 - 2.*c1 + c2),
            (c0 - c1 - root) / (c0 - 2.*c1 + c2)]



#==============================================================
#                        class StrokeLeg
#==============================================================

class StrokeEndError(Exception):
    def __init__(self):
        pass

class StrokeLeg(object):
    """Given a segment (a "leg") of a stroke of a path, record:
    - bc:            BCurve.BezierCurve (the same stroke in another format)
    - chop_paths:   [gimp.Vectors] (bc chopped at its anchors)
    - chop_lengths   [float] (lengths of paths in chop_paths along the curve,
                               in the same order as in bc)
    - start:         complex (starting point of the leg)
    - end:           complex (ending point of the leg)
    - length:        float (length of the leg)
    - distance:      float (the length of the curve from its start to leg's start)
    - at_stroke_end: boolean (input True if you know this to be the end of the stroke)
    Raise StrokeEndError if there is no room on the stroke, except if
    the caller sets at_stroke_end=True.
    All points are complex numbers (the plane is viewed as the complex number plane).
    """
    #def __init__(self, vectors, stroke, distance, length, at_stroke_end=False):
    #def __init__(self, image, vectors, stroke, distance, length, at_stroke_end=False):
    def __init__(self,
                 image,
                 vectors,
                 stroke,
                 bc,
                 chop_paths,
                 chop_lengths,
                 distance,
                 length,
                 at_stroke_end=False):
        PRECISION = 1e-10
        #self.path = vectors
        #self.stroke = stroke
        self.length = length
        self.distance = distance
        self.at_stroke_end = at_stroke_end
        self.bc = bc
        
        self.chop_paths = chop_paths
        self.chop_lengths = chop_lengths
        
        x0, y0, slope0, valid0 = pdb.gimp_vectors_stroke_get_point_at_dist(
                      vectors, stroke.ID, distance, PRECISION)
        x1, y1, slope1, valid1 = pdb.gimp_vectors_stroke_get_point_at_dist(
                      vectors, stroke.ID, distance+length, PRECISION)
        if not valid0:
            raise StrokeEndError()
        self.start = complex(x0,y0)
        if valid1:
            self.end = complex(x1,y1)
        else:
            if at_stroke_end: # We know we are adding the end of the stroke
                x1,y1 = stroke.points[0][-4:-2]
                self.at_stroke_end = True
                self.end = complex(x1,y1)
                # Do not raise StrokeEndError
            else:
                stroke_length = pdb.gimp_vectors_stroke_get_length(vectors, stroke.ID, PRECISION)
                if stroke_length - (distance+length) < PRECISION: # At stroke end
                    x1,y1 = stroke.points[0][-4:-2]
                    self.at_stroke_end = True
                    self.end = complex(x1,y1)
                raise StrokeEndError()
    def __str__(self):
        s = "StrokeLeg"
        #s += "\n  path:   "+str(self.path.name)
        #s += "\n  stroke: "+str(self.stroke.ID)
        s += "\n  start = "+str(self.start)
        s += "\n  end   = "+str(self.end)
        s += "\n  distance = "+str(self.distance)
        s += "\n  length   = "+str(self.length)
        #s += "\n  leg.bc:"
        #s += str(self.bc)
        s += "\n  chop_path stroke lengths:"
        s += "\n    "+str(self.chop_lengths)
        return s

# ------------
# chop_stroke0
# ------------
# Chop the first stroke of path at anchors.
# Return
# - the first stroke as BCurve.BezierCurve;
# - the chopped parts as a list of paths [gimp.Vectors];
# - the lengths of the chopped parts (measured along the curve).
# Note: Consequently, each path in the returned list of paths consists of
# one stroke and the stroke consists of one Bezier arc.
# Args:
# - image
# - path: gimp.Vectors
# Returns:
# - BCurve.BezierCurve
# - [gimp.Vectors]
# - [float]
def chop_stroke0(image, path):
    PRECISION = 1e-10
    gv = vectors_object2gv(path)
    bc = gv.stroke_list[0].gs2bc() # First stroke only
    chop_paths = []
    chop_lengths = []
    for index,ba in enumerate(bc.bezier_arcs):
        chop_bc = BCurve.BezierCurve(bezier_arcs = [ba])
        chop_paths.append(chop_bc.bc2vectors_object(image, name='chop '+str(index)))
        
        chop_gs = chop_bc.bc2gs()
        one_gv = BCurve.GimpVectors(stroke_list = [chop_gs],
                                    name = 'chop '+str(index)) # Not needed?
        one_path = one_gv.gv2vectors_object(image)
        one_length = pdb.gimp_vectors_stroke_get_length(one_path,
                                                        one_path.strokes[0].ID,
                                                        PRECISION)
        chop_lengths.append(one_length)
    return bc, chop_paths, chop_lengths


# -------------------------
# strokes_to_legs_by_length
# -------------------------
# Given a path 'target_path', extract from its strokes 'nlegs' legs where the first
# has length=leg_length, the second kx*leg_length, the third
# kx*kx*leg_length, and so on, where kx = 1-tapering.
# The arguments 'left_overshoot' and 'right_overshoot' tell how much the legs
# should cover from each stroke of 'target_path' before the start and beyond the end
# - which means that we construct extended versions of the strokes.
# Args:
# - image
# - target_path:      gimp.Vectors
# - leg_length:      float
# - nlegs:           int
# - left_overshoot:  float
# - right_overshoot: float
# - tapering:        float
# Returns:
# - [[StrokeLeg]] (Each inner list corresponds to one stroke)
def strokes_to_legs_by_length(image,
                              target_path,
                              leg_length,
                              nlegs,
                              left_overshoot,  # for Source in standard position
                              right_overshoot, # for Source in standard position
                              tapering):
    PRECISION = 1e-10
    global MAX_NCOPIES
    kx = 1-tapering
    leglist_list = []
    for stroke_count, stroke in enumerate(target_path.strokes):
        stroke_length = pdb.gimp_vectors_stroke_get_length(target_path, stroke.ID,
                                                           PRECISION)
        # Find:
        # - the actual number of legs to make;
        # - the length of the last leg to make:
        if tapering == 0:
            if nlegs > 0:
                actual_nlegs = min(nlegs, int(stroke_length / leg_length))
            else:
                actual_nlegs = int(stroke_length / leg_length)
            last_leg_length = leg_length
        else:
            geom_sum = 0
            last_term = leg_length
            if nlegs == 0:
                nlegs = MAX_NCOPIES
            for i in range(nlegs):
                geom_sum = kx * geom_sum + leg_length
                if geom_sum > stroke_length:
                    actual_nlegs = i
                    last_leg_length = last_term
                    break
                last_term *= kx
            else: # No break, so:
                actual_nlegs = nlegs
                last_leg_length = last_term
        ext_path, ext_left, ext_right = extend_path_stroke_with_straight(image,
                           target_path,
                           stroke,
                           left_overshoot * leg_length,
                           right_overshoot * last_leg_length,
                           extensions_absolute=True)
        ext_path.name = target_path.name+' extended '+str(stroke_count)
        leglist = []
        leg_start = ext_left
        current_length = leg_length
        
        ext_bc, chop_paths, chop_lengths = chop_stroke0(image, ext_path)
        
        for i in range(actual_nlegs):
            try:
                #leglist.append(StrokeLeg(vectors  = ext_path,
                leglist.append(StrokeLeg(image,
                                         vectors  = ext_path,
                                         stroke   = ext_path.strokes[0],
                                         bc       = ext_bc,
                                         chop_paths = chop_paths,
                                         chop_lengths = chop_lengths,
                                         distance = leg_start,
                                         length   = current_length))
                leg_start += current_length
                current_length *= kx
            except StrokeEndError:
                break
            if len(leglist) == nlegs:
                break
        leglist_list.append(leglist)
    #print("strokes_to_legs_by_length: Returning legs:")
    #for i,leglist in enumerate(leglist_list):
    #    print("List for stroke "+str(i))
    #    for j,leg in enumerate(leglist):
    #        print("leg "+str(j))
    #        print(str(leg))
    return leglist_list

# --------------------------
# strokes_to_legs_by_ncopies
# --------------------------
# Given a path 'target_path', extract from its strokes 'nlegs' legs where the first
# has length=l, the second kx*l, the third kx*kx*, and so on, where kx = 1-tapering
# and l is determined from the requirement that the whole strokes must be filled.
# The arguments 'left_overshoot' and 'right_overshoot' tell how much the legs
# should cover from each stroke of 'target_path' before the start and beyond the end
# - which means that we construct extended versions of the strokes.
# Args:
# - image
# - target_path:      gimp.Vectors
# - nlegs:           integer
# - left_overshoot:  float
# - right_overshoot: float
# - tapering:        float
# Returns:
# - [[StrokeLeg]] (Each inner list corresponds to one stroke)
def strokes_to_legs_by_ncopies(image,
                               target_path,
                               nlegs,
                               left_overshoot,  # for Source in standard position
                               right_overshoot, # for Source in standard position
                               tapering):
    kx = 1-tapering
    MAX_ROUNDS = 100
    PRECISION = 1e-10
    try:
        geom_sum = (1 - kx**nlegs) / (1-kx)
    except ZeroDivisionError:
        geom_sum = nlegs
    leglist_list = []
    for stroke_count, stroke in enumerate(target_path.strokes):
        stroke_length = pdb.gimp_vectors_stroke_get_length(target_path, stroke.ID,
                                                           PRECISION)
        leg_length = stroke_length / geom_sum
        # Find the length of the last leg to be added:
        if tapering == 0:
            last_leg_length = leg_length
        else:
            last_leg_length = leg_length * (kx**(nlegs-1))
        ext_path, ext_left, ext_right = extend_path_stroke_with_straight(image,
                           target_path,
                           stroke,
                           left_overshoot * leg_length,
                           right_overshoot * last_leg_length,
                           extensions_absolute=True)
        ext_path.name = target_path.name+' extended '+str(stroke_count)
        leglist = []
        leg_start = ext_left
        current_length = leg_length
        
        ext_bc, chop_paths, chop_lengths = chop_stroke0(image, ext_path)
        
        for i in range(nlegs):
            try:
                #leglist.append(StrokeLeg(vectors  = ext_path,
                leglist.append(StrokeLeg(image,
                                         vectors  = ext_path,
                                         stroke   = ext_path.strokes[0],
                                         bc       = ext_bc,
                                         chop_paths = chop_paths,
                                         chop_lengths = chop_lengths,
                                         distance = leg_start,
                                         length   = current_length))
                leg_start += current_length
                current_length *= kx
            except StrokeEndError:
                break
        #else:
        #    raise Exception("strokes_to_legs_by_ncopies: Reached MAX_ROUNDS = "+str(MAX_ROUNDS))
        if len(leglist) == nlegs-1:
            # Can this happen: can the last piece be left out?
            # Just to be certain we add it if needed:
            #print("Add last leg especially.")
            #leglist.append(StrokeLeg(target_path, stroke, stroke_length-current_length,
            leglist.append(StrokeLeg(image,
                                     #target_path, # VIRHE??
                                     #stroke,      # VIRHE??
                                     #stroke_length-current_length,
                                     #current_length, at_stroke_end=True))
                                     vectors  = ext_path,
                                     stroke   = ext_path.strokes[0],
                                     bc       = ext_bc,
                                     chop_paths = chop_paths,
                                     chop_lengths = chop_lengths,
                                     #distance = stroke_length-current_length,
                                     distance = leg_start, # Bug fix 0.8
                                     length   = current_length,
                                     at_stroke_end=True))
        else: # Check if the last end is precisely the last anchor:
            last_anchor_v = stroke.points[0][-4:-2]
            last_anchor = complex(last_anchor_v[0],last_anchor_v[1])
            leglist[-1].end = last_anchor
            leglist[-1].at_stroke_end=True
        leglist_list.append(leglist)
    return leglist_list

#==============================================================
#                Simple Bezier arc
#==============================================================
# ----------
# simple_arc
# ----------
# This is an ad hoc routine to draw simple short arcs.
# Given points A,B and in then unit direction vectors as tangent directions,
# make a simple Bezier arc from A to B.
# Assumes that A != B.
# Args:
# - A, B: complex
# - Adir, Bdir: complex (should point in the direction of the arc)
# Returns:
# - [complex,complex,complex,complex] (control points)
def simple_arc(A, B, Adir, Bdir):
    chord = abs(A-B)
    if chord == 0:
        #raise Exception("simple_arc: equal end points")
        return [A,A,B,B]
    return [A, A + Adir * chord / 4, B + Bdir * chord / 4, B]


#==============================================================
#                Bezier arc from three tangents
#==============================================================
# --------------
# half_hausdorff
# --------------
# The error measure: how much does the Bezier arc (represented by cp4)
# deviate from the given sample points?
# Return the absolute error.
# Args:
# - cp4:     [complex,complex,complex,complex] (control points)
# - samples: [complex]
# - level:   integer 0..10 (working level)
# Returns:
# - float (absolute error)
def half_hausdorff(cp4, samples, level=5):
    max_error = 0
    for v in samples:
        _,w = bezier_closest_point(cp4, v, restrict01=True)
        max_error = max(abs(v-w), max_error)
    return max_error

# --------------------------
# bezier_from_three_tangents
# --------------------------
# Find control points for a Bezier curve B(t), given
# - end points (= B(0) and B(1));
# - one third point;
# - tangent directions in those three points (assuming that none of the three
#   points will turn out to be a cusp point).
# Args:
# - p0, p3, p:        complex
# - u0, u3, u:        complex
# Returns:
# - list [ [[q0,q1,q2,q3],t,check],  [[q0,q1,q2,q3],t,check],  ... ]
#   containing zero to three items, where
#   - [q0,q1,q2,q3] is the list of control points of a Bezier curve B(t) such that
#     - the points q0,q3 are the given p0,p3;
#     - B(t)=p for some t;
#     - B'(0), B'(1), B'(t) are scalar multiples of u0,u3,u, respectively
#       (so that if they are not cusp points, then the curve B(t) has tangents
#       parallel to the prescribed u0,u3,u0, respectively;
#   - t is the parameter value where the curve reaches the point p, i.e., B(t)=p;
#   - check is [item,item,item] where each 'item' corresponds to
#     p0, p3, or p, and 'item' is None or True or False, meaning:
#     - None:  cusp;
#     - True:  B' at that point points in the same direction as the
#              prescribed u0,u3, or u;
#     - False: B' at that point points in the opposite direction from the
#              prescribed u0,u3, or u
# Exceptions:
# - If inputs are not ok, raises Bezier3TangentsInputError.
# Notes:
# 1. If the returned Bezier arc has a cusp in one of the points p0,p3,p, then
#    its tangent at that point has NOT the required direction but is something
#    one cannot even guess!
# 2. Solutions with t=0 or t=1 are rejected.
# 3. The plane is viewed as the complex number plane,
#    so all points are complex numbers.
def bezier_from_three_tangents(p0, p3, p, u0, u3, u):
    from math import sqrt
    ZERO = 1e-8
    bernstein = [lambda x: (1-x)**3,
                 lambda x: 3*(1-x)*(1-x)*x,
                 lambda x: 3*(1-x)*x*x,
                 lambda x: x**3]
    def Q(t,z0,z3):
        coeff0 = -(bernstein[0](t) + bernstein[1](t))
        coeff3 =   bernstein[2](t) + bernstein[3](t)
        return coeff0 * z0 + coeff3 * z3
    def cross(v,w):
        return -v.real*w.imag + v.imag*w.real
    def check_directions(v0,v3,v, control_points, t0,t3,t):
        epsilon = 1e-10
        nv0,_ = bezier_unit_direction_vector(t0, control_points)
        nv3,_ = bezier_unit_direction_vector(t3, control_points)
        nv,_  = bezier_unit_direction_vector(t,  control_points)
        directions = []
        num = 0
        for x,y in [[v0,nv0], [v3,nv3], [v,nv]]:
            num += 1
            if -epsilon < x.real*y.real + x.imag*y.imag < epsilon: # cusp
                directions.append(None)
            elif x.real*y.real + x.imag*y.imag < 0:  # opposite direction
                directions.append(False)
            else:
                directions.append(True)
        return directions
    def non_zero(v):
        epsilon = 1e-14
        return abs(v)**2 > epsilon
    def parallel(v,w):
        epsilon = 1e-12
        len_v = abs(v)
        len_w = abs(w)
        s = cross(v,w)
        return -epsilon < s / (len_v*len_w) < epsilon
    C0 = cross(u3,u)
    C3 = cross(u,u0)
    C  = cross(u0,u3)
    z0 = -p0+p     # -p0 + p
    z3 = -p+p3     # -p  + p3
    z  = -p3+p0   # -p3 + p0
    # Check input
    if not (non_zero(z0) and non_zero(z3)): # p=p0 or p=p3
        m = 'bezier_from_three_tangents: two points too close (identical?)'
        raise Bezier3TangentsInputError(m, code=0)
    if not (non_zero(u0) and non_zero(u3) and non_zero(u)): # u0,u3, or u is 0
        m = 'bezier_from_three_tangents: a tangent vector is 0 (or too close)'
        raise Bezier3TangentsInputError(m, code=1)
    if parallel(u0,u3) and parallel(u,u3) and parallel(u,u0): # u0,u3,u parallel
        # Check if we have a linear tangent triangle:  p0---p---p3
        if abs(abs(-p0+p) + abs(-p+p3) - abs(-p0+p3)) < ZERO:
            if parallel(-p0+p3, u):
                cp4 = [p0,p0,p3,p3]
                p03 = -p0+p3
                coeff = [vdot(p0-p,p03), vdot(p0-p,p03), vdot(p3-p,p03), vdot(p3-p,p03)]
                poly_zeroes = zeroes_of_bernstein_polynomial(coeff)
                if poly_zeroes is None:
                    m = "Linear tangent triangle, apex not on the base line"
                    raise Bezier3TangentsInputError(m, code=2)
                ts = [t for t in poly_zeroes[0] if 0 < t < 1]
                if len(ts) != 1:
                    m = "Linear tangent triangle, len(ts) = "+str(len(ts))
                    raise Bezier3TangentsInputError(m, code=3)
                t = ts[0]
                check = check_directions(u0,u3,u, cp4, 0.,1.,t)
                return [[cp4, t, check]]
        m = 'bezier_from_three_tangents: all tangents parallel (or too close to be)'
        raise Bezier3TangentsInputError(m, code=3)
    #
    z0xC0u0 = C0*cross(z0,u0)
    z0xCu   = C *cross(z0,u)
    z3xC3u3 = C3*cross(z3,u3)
    z3xCu   = C *cross(z3,u)
    #
    # Coefficient of H(t) in Bernstein basis:
    c0 = -z0xC0u0 + z0xCu
    c1 = -z0xC0u0
    c2 = -z3xC3u3
    c3 = -z3xC3u3 + z3xCu
    #
    polynomial_zeroes = zeroes_of_bernstein_polynomial([c0,c1,c2,c3])
    if polynomial_zeroes is None: # Polynomial identically zero: raise error
        m = 'bezier_from_three_tangents: polynomial identically 0'
        raise Bezier3TangentsInputError(m, code=5)
    zeroes, extrema, inflection = polynomial_zeroes
    results = []
    for t in zeroes:
        if t in (0,1): # Reject
            continue
        Qt = Q(t,z0,z3)
        q0t = -(t**3)*z + (2-3*t)*z0
        q3t = ((1-t)**3)*z + (1-3*t)*z3
        if abs(C) >= abs(C3):
            alpha = -cross(Qt,u3) / (bernstein[1](t) * C)
        else:
            alpha = -cross(q0t,u) / (bernstein[1](t) * C3)
        if abs(C) >= abs(C0):
            beta = -cross(Qt,u0) / (bernstein[2](t) * C)
        else:
            beta = -cross(q3t,u) / (bernstein[2](t) * C0)
        q0 = p0
        q1 = p0 + alpha * u0
        q2 = p3 - beta * u3
        q3 = p3
        check = check_directions(u0,u3,u, [q0,q1,q2,q3], 0.,1.,t)
        results.append([[q0,q1,q2,q3], t, check])
    return results

# ------------------------------
# bezier_arc_from_three_tangents
# ------------------------------
# Version of bezier_from_three_tangents where accepted are only answers that
# serve building a Bezier arc from the data:
# - The directions of B' at the end points must be u0 and u3
#   (opposite directions not accepted);
# - the paremeter t (giving p) must satisfy 0<t<1 so that p will be on the arc
#   (not on its extension).
# Args:
# - p0, p3, p: complex
# - u0, u3, u: complex
# Returns:
# - [[cp4, t]] where cp4 is the control point quadruple and t is the parameter
#   value with which p is reached.
# Note: The returned list will be empty if no accepted arc was found.
# Generally there are 0-3 answers. (But can there be 2 or 3?)
def bezier_arc_from_three_tangents(p0, p3, p, u0, u3, u):
    results = bezier_from_three_tangents(p0, p3, p, u0, u3, u)
    # Accept only those with 0<t<1:
    results = [r for r in results if 0 < r[1] < 1]
    # Accept only those where the directions of B' at the end points are as required:
    results_accept = [r for r in results if (r[2][0] is True) and (r[2][1] is True)]
    # Drop the direction checkings:
    return [r[:-1] for r in results_accept]

class Bezier3TangentsInputError(Exception):
    def __init__(self, message, code=None): # 'message' is a string
        self.code = code
        self.message = message
    def __str__(self):
        m = "Bezier3TangentsInputError: Code "+str(self.code)
        m += "\n"+self.message
        return m


#==============================================================
#    Copy a source pattern along a path: the transformation
#==============================================================
# ------------------------
# bcurve_get_point_at_dist
# ------------------------
# Routine to replace pdb.gimp_vectors_stroke_get_point_at_dist().
# Given
# - a Bezier curve bc:BCurve.BezierCurve;
# - the same chopped at its anchors, as a list of paths with 4 control points each;
# - the lengths of the chopped paths.
# So, in the pathlist
# - each path consists of one stroke and the stroke of one segment (4 control points);
# - the segments are butting.
# Find the point X at the given distance 'point_distance' from the start of bc.
# Return the point, the Bezier arc where X was found, and the
# parameter value for X in the Bezier arc (4 control points).
# Args:
# - bc:             BCurve.BezierCurve
# - pathlist:       [gimp.Vectors]
# - arc_lengths:    [float]
# - point_distance: float
# Returns:
# - complex          (the found point X)
# - BCurve.BezierArc (the Bezier arc where X was found)
# - float            (the parameter value for X on that arc)
# - boolean          ('valid': True if X was found)
# Note: This replaces pdb.gimp_vectors_stroke_get_point_at_dist() for efficiency
# rea: now the pdb procedure is applied to one Bezier arc at
# a time, not to the whole bc.
def bcurve_get_point_at_dist(bc,
                             pathlist,
                             arc_lengths,
                             point_distance,
                             print_info=False):
    PRECISION = 1e-10
    if point_distance < 0:
        return None, None, None, False
    total_length = 0
    arc_distance = 0 # The distance of the start of the arc
    for i,arc_length in enumerate(arc_lengths):
        arc_distance = total_length
        total_length += arc_lengths[i]
        if total_length >= point_distance:
            arc_index = i
            break
    else: # No break
        return None, None, None, False
    path = pathlist[arc_index] # arc_index is the index of the Bezier arc
    x, y, slope, valid = pdb.gimp_vectors_stroke_get_point_at_dist(
                  path,
                  path.strokes[0].ID,
                  point_distance - arc_distance,
                  PRECISION)
    if (not valid) or (point_distance < 0):
        stroke_length = pdb.gimp_vectors_stroke_get_length(path, stroke.ID, PRECISION)
        if print_info:
            print("map_point_direction_along_leg: Error:")
            if not valid:
                print("not valid")
            else:
                print("point_distance negative: "+str(point_distance))
            print("leg.distance = "+str(leg.distance))
            print("point = "+str(point))
            print("leg.length = "+str(leg.length))
            print("point_distance = "+str(point_distance))
            print("stroke_length  = "+str(stroke_length))
            print("leg:")
            print(str(leg))
        m = "pdb.gimp_vectors_stroke_get_point_at_dist returned 'not valid'."
        m += "\nThis should not happen. This is a bug."
        raise Exception(m)
    # The point found by pdb.gimp_vectors_stroke_get_point_at_dist:
    pdb_point = complex(x,y)
    # This point is inaccurate, off the curve(?). So is slope(??).
    # We try to find better.
    ba = bc.bezier_arcs[arc_index]
    t,B = bezier_closest_point(ba.cp4, pdb_point, restrict01=True)
    Bt = bezier_rv(t, ba.cp4)
    if print_info:
        print("bcurve_get_point_at_dist: error = "+str(abs(Bt-pdb_point)))
    return Bt, ba, t, True

# -----------------------------
# map_point_direction_along_leg
# -----------------------------
# Given a point and a direction (presumably a point on some curve and
# the tangent direction in that point), map them in the hopeful transformation
# onto a stroke leg. Assume that the input point is in a "standard position":
# the "base" is already mapped onto the line segment [0,1].
# If the input direction is None, the output direction will be None.
# Args:
# - point:      complex
# - direction:  complex
# - leg:        StrokeLeg
# - tapering:   float
# - print_info: boolean
# Returns:
# - complex (mapped point)
# - complex (mapped direction, a unit vector) or None
def map_point_direction_along_leg(point, direction, leg, tapering, print_info=False):
    if print_info:
        print("\nmap_point_direction_along_leg: Starting with")
        print("point = "+str(point))
        print("direction = "+str(direction))
    PRECISION = 1e-10
    c = 1-tapering
    pointx, pointy = point.real, point.imag
    try:
        fx = (1 - (c**pointx)) / (1-c)
        fdotx = -log(c)*(c**pointx) / (1-c)
    except ZeroDivisionError:
        fx = pointx
        fdotx = 1
    point_distance = leg.distance + fx * leg.length
    best_point, best_ba, best_t, valid = bcurve_get_point_at_dist(
                                                   leg.bc,
                                                   leg.chop_paths,
                                                   leg.chop_lengths,
                                                   point_distance)
    #
    s, s_case = bezier_unit_direction_vector(best_t, best_ba.cp4)
    if s is None:
        raise Exception("Constant?")
    if s_case == 2 and best_t == 1: # cusp at the end point: opposite direction
        s = -s
    sperp = complex(-s.imag, s.real)
    # The point transformed:
    new_point = best_point + leg.length * (pointy - (1-c)*fx*pointy) * sperp
    # The direction transformed:
    if direction is None:
        new_direction = None
    else:
        d1,d2 = direction.real, direction.imag
        new_direction = (
           fdotx * d1 * s
           + (d2 - (1-c) * (fx * d2 + fdotx * d1 * pointy)) * sperp
           )
        # An additional term:
        try:
            Adot = bezier_dot_rv(best_t, best_ba.cp4)
            s2 = bezier_dot_dot_rv(best_t, best_ba.cp4) / (abs(Adot)**2)
            new_direction -= (leg.length * fdotx * d1
                                         * (pointy - (1-c) * fx * pointy)
                                         * vdot(sperp,s2)
                              ) * s
        except ZeroDivisionError:
            pass
        new_direction = new_direction/abs(new_direction)
    if print_info:
        print("map_point_direction_along_leg: Returning:")
        print("new_point = "+str(new_point))
        print("new_direction = "+str(new_direction))
    return new_point, new_direction

#==============================================================
#   Copy a source pattern along a path: high level routines
#==============================================================

# -----------------
# preprocess_bclist
# -----------------
# Given bclist:[BCurve.BezierCurve] find inflection points and add anchors at
# those points.
# Args:
# - bclist:[BCurve.BezierCurve]
# Returns:
# - [BCurve.BezierCurve]
def preprocess_bclist(bclist):
    new_bclist = []
    for bc in bclist:
        bas = completed_bezier_arcs(bc) # Handles closed cases ok: closes the gap.
        if (len(bc.bezier_arcs) == 1) and (len(bc.bezier_arcs[0].cp4) == 1):
            # 1-anchor stroke
            new_bclist.append(deepcopy(bc))
            continue
        new_balist = []
        for ba in bas:
            special = bezier_special_points(ba.cp4)
            if special[0] in (1,2): # Case: inflection points
                ts = special[1]     # Parameter values
                ts = [t for t in ts if 0 < t < 1]
                if len(ts) == 0: # No inflection points in the arc: copy ba as such
                    new_balist.append(deepcopy(ba))
                else: # Subdivide
                    subdiv_cp4 = bezier_subdivide_at_parameters(ts, ba.cp4) # [cp4]
                    new_balist += [BCurve.BezierArc(cp4=cp4) for cp4 in subdiv_cp4]
            else: # No inflection points: copy ba as such
                new_balist.append(deepcopy(ba))
        if bc.closed: # Remove the gap-closing arc
            gap_ba = new_balist[-1]
            new_balist = new_balist[:-1]
            new_bc = BCurve.BezierCurve(bezier_arcs=new_balist,
                                        head_handle=gap_ba.cp4[2],
                                        tail_handle=gap_ba.cp4[1],
                                        closed=True
                                        )
        else:
            new_bc = BCurve.BezierCurve(bezier_arcs=new_balist)
        new_bclist.append(new_bc)
    return new_bclist

# ----------------------
# preprocess_target_path
# ----------------------
# Currently: 
# 1. Remove open 1-anchor strokes.
# 2. Make closed strokes open, adding the gap-closing arc.
# Args:
# - image
# - path: gimp.Vectors
# Returns:
# - gimp.Vectors
def preprocess_target_path(image, path):
    gv = vectors_object2gv(path)
    new_stroke_list = []
    for gs in gv.stroke_list:
        if gs.closed:
            new_balist = completed_bezier_arcs(gs.gs2bc())
            new_bc = BCurve.BezierCurve(bezier_arcs = new_balist) # Made open
            new_stroke_list.append(new_bc.bc2gs())
        elif len(gs.cp_list) <= 3: # Open 1-anchor stroke
            pass
        else:
            new_stroke_list.append(deepcopy(gs))
    new_gv = BCurve.GimpVectors(stroke_list = new_stroke_list)
    new_gv.name = path.name+' preprocessed'
    return new_gv.gv2vectors_object(image)

# ----------------------
# cp4_to_cp4list_distort
# ----------------------
# Copies cp4 onto the stroke leg, suitably scaled and rotated and distorted.
# The points 0 and 1 are mapped onto the leg end points.
# Returns list of cp4's and a relative error measure.
# Args:
# - cp4:        [complex,complex,complex,complex] (control point quaduple)
# - leg:        StrokeLeg
# - tapering:   float
# - print_info: boolean
# Returns:
# - [[complex,complex,complex,complex]]
# - float (error)
def cp4_to_cp4list_distort(cp4, leg, tapering, print_info=False):
    def wild_handles(cp4):
        WILD = 100 # ad hoc
        p0,p1,p2,p3 = cp4
        if print_info:
            if (abs(-p0+p1)+abs(-p3+p2) > WILD * abs(-p0+p3)):
                print("WILD cp4_approx:")
                for cp in cp4:
                    print("    "+str(cp))
        return (abs(-p0+p1)+abs(-p3+p2) > WILD * abs(-p0+p3))
    if print_info:
        print("\n==========  Starting cp4_to_cp4list_distort  ===========")
        print("with p0,p3 = "+str((cp4[0],cp4[-1])))
        print("and leg ends: "+str((leg.start, leg.end)))
    ZERO = 1e-8
    MAX_ROUNDS = 100
    MAX_LEVEL_ALLOWED = 12 # level of subdivision
    MAX_ERROR_ALLOWED = 0.005
    SHORT_ARC = 5 # pixels
    PARALLEL_TANGENTS = 1e-8
    unfinished_cp4list = [(cp4,0)] # [(cp4,subdiv_depth)]
    finished_cp4list = []          # [(cp4,error)]
    # Take from unfinished_cp4list one cp4 at a time,
    # try to transform it, and if successful, put the result in finished_cp4list;
    # otherwise subdivide the cp4 and put the pieces back in unfinished_cp4list.
    #
    # Transformation phases:
    # 1. Do the transformation:
    # 2. Construct an approximate Bezier arc.
    # 3. Check the success of the approximation. If failed, do subdivision and
    #    jump to next round in the loop. Otherwise (success), store.
    subdiv_depth_max = 0 # Just to record deepest subdivision for development
    for round_count in range(MAX_ROUNDS):
        if print_info:
            print("--------------  Round "+str(round_count)+"  ------------------")
        try:
            cp4_under_work, subdiv_depth = unfinished_cp4list.pop(0) # First unfinished arc
        except IndexError:
            if print_info:
                print("unfinished_cp4list is exhausted")
                print("Break at round "+str(round_count))
            break # unfinished_cp4list is empty
        cp4_0 = cp4_under_work
        # Vertical lines are mapped to straight lines: no need for subdivision
        # or approximation! Check first this.
        x0,x1,x2,x3 = [z.real for z in cp4_0]
        if (abs(-x0+x1) < ZERO) and (abs(-x0+x2) < ZERO) and (abs(-x0+x3) < ZERO):
            if print_info:
                print("Vertical line")
            p0,p1,p2,p3 = cp4_0
            try:
                p0_1,_ = map_point_direction_along_leg(p0, None, leg, tapering)
                p3_1,_ = map_point_direction_along_leg(p3, None, leg, tapering)
            except StrokeEndError:
                if print_info:
                    print("StrokeEndError from p0_1, p0_3")
                    print("Break at round "+str(round_count))
                break
            cp4_final = [p0_1, p0_1, p3_1, p3_1]
            finished_cp4list.append((cp4_final,0)) # set error = 0
            if print_info:
                print("Append to finished_cp4list")
            continue
        # Phase 1: Transformation
        C0,D0 = cp4_0[0], cp4_0[-1]
        end_dirs = bezier_arc_end_directions(cp4_0)
        if end_dirs is None: # One-point arc?
            if print_info:
                print("end_dirs is None")
            continue
        C0direction, D0direction = bezier_arc_end_directions(cp4_0)
        try:
            C1, C1direction = map_point_direction_along_leg(C0, C0direction, leg, tapering)
            D1, D1direction = map_point_direction_along_leg(D0, D0direction, leg, tapering)
        except StrokeEndError:
            if print_info:
                print("StrokeEndError from C1, D1:")
                print("C0, D0 = "+str((C0,D0)))
                print("Break at round "+str(round_count))
            break
        # If the arc is very short we use a simple solution:
        if abs(C1-D1) < SHORT_ARC:
            cp4_simple = simple_arc(C1, D1, C1direction, D1direction)
            finished_cp4list.append((cp4_simple,0)) # set error = 0
            if print_info:
                print("C1,D1 = "+str((C1,D1)))
                print("abs(C1-D1) = "+str(abs(C1-D1))+": Use a simple arc.")
                print("Append to finished_cp4list")
            continue
        # If the arc is very flat (looking at the end points only)
        # we use a simple solution:
        if abs(C1direction + D1direction) < PARALLEL_TANGENTS: ###########  KOE
            cp4_simple = simple_arc(C1, D1, C1direction, D1direction)
            finished_cp4list.append((cp4_simple,0)) # set error = 0
            if print_info:
                print("C1,D1 = "+str((C1,D1)))
                print("abs(C1-D1) = "+str(abs(C1-D1)))
                print("abs(C1direction - D1direction) = "+str(abs(C1direction-D1direction))+": Use a simple arc.")
                print("Append to finished_cp4list")
            continue
        # Get sample points from the mathematically correct curve
        # for later checking of approximation error.
        # First: Take sample points from cp4_0:
        try:
            t_point_side_extremals = bezier_side_extremals(cp4_0, level=1)
            t_point_left, t_point_middle, t_point_right = t_point_side_extremals
            t_mean,P0 = t_point_middle
            t_left, P0_left = t_point_left
            t_right, P0_right = t_point_right
        except Exception as e:
            if print_info:
                print("bezier_side_extremals raised:")
                print(str(e))
            t_mean,P0 = 0.5, bezier_rv(0.5, cp4_0)
            t_left, P0_left = 0.25, bezier_rv(0.25, cp4_0)
            t_right, P0_right = 0.75, bezier_rv(0.75, cp4_0)
        # Second: Transform the sample points:
        try:
            P0direction = bezier_dot_rv(t_mean, cp4_0)
            P0direction = P0direction / abs(P0direction)
        except ZeroDivisinError:
            raise Exception("cp4_to_cp4list_distort: P0direction = 0 - a cusp?")
        P1,P1direction = map_point_direction_along_leg(P0, P0direction, leg, tapering)
        P1_left,_ = map_point_direction_along_leg(P0_left, None, leg, tapering)
        P1_right,_ = map_point_direction_along_leg(P0_right, None, leg, tapering)
        # Phase 2: Construct Bezier approximation:
        try:
            arc_data = bezier_arc_from_three_tangents(C1, D1, P1,
                                           C1direction, -D1direction, P1direction)
        except Bezier3TangentsInputError as e: # Failure in approximation
            if print_info:
                print("Bezier3TangentsInputError "+ str(e))
                print("C1,P1,D1 = "+str((C1,P1,D1)))
                print("Direction vectors = "+str((C1direction, P1direction, -D1direction)))
                print("abs(C1-D1) = "+str(abs(C1-D1)))
                print("abs(C1direction + D1direction) = "+str(abs(C1direction + D1direction)))
                print("C0,P0,D0 = "+str((C0,P0,D0)))
                print("Direction vectors = "+str((C0direction, P0direction, -D0direction)))
            if subdiv_depth < MAX_LEVEL_ALLOWED:
                if print_info:
                    print("Subdivide")
                subdiv = bezier_subdivide_at_parameters([t_mean], cp4_under_work)
                subdiv_with_level = [(sd, subdiv_depth+1) for sd in subdiv]
                unfinished_cp4list = subdiv_with_level + unfinished_cp4list
                subdiv_depth_max = max(subdiv_depth_max, subdiv_depth+1)
            else:
                if print_info:
                    print("Reached MAX_LEVEL_ALLOWED = "+str(MAX_LEVEL_ALLOWED))
                    print("Append to finished_cp4list")
                # Put to finished though not good.
                finished_cp4list.append((cp4_approx, error_approx))
            continue
        if len(arc_data) == 0:
            if print_info:
                print("len(arc_data) == 0: simple work-around")
            # Make a simple workaround
            chord = abs(C1-D1)
            try:
                q1 = C1 + (C1direction/abs(C1direction)) * chord/4
            except ZeroDivisionError:
                q1 = C1
            try:
                q2 = D1 + (D1direction/abs(D1direction)) * chord/4
            except ZeroDivisionError:
                q2 = D1
            cp4_approx = [C1,q1,q2,D1]
            t_approx = 0.5
        elif len(arc_data) == 1:
            cp4_approx, t_approx = arc_data[0]
        else: # len(arc_data) = 2 or 3: got 2 or 3 acceptable Bezier arcs
            if print_info:
                print("len(arc_data) > 1: ad-hoc choose one")
            # A very simple ad-hoc way to choose one:
            # Choose the one where t is closest to 1/2.
            # For later development: Should choose the one which runs closest
            # to the desired curve. Should compute some sample points etc.
            # (But it seems to make no big differense.)
            cp4_approx, t_approx = min(arc_data, key=(lambda ad:abs(ad[1]-0.5)))
        # Phase 3: Check the approximation
        samples = [P1_left, P1_right]
        
        #################  More sample points?
        ## Get four more sample points
        #t_left_mean = (t_left + t_mean)/2
        #t_right_mean = (t_right + t_mean)/2
        #t_left_left = t_left/2
        #t_right_right = t_right/2
        #P0_left_mean = bezier_rv(t_left_mean, cp4_0)
        #P0_right_mean = bezier_rv(t_right_mean, cp4_0)
        #P0_left_left = bezier_rv(t_left_left, cp4_0)
        #P0_right_right = bezier_rv(t_right_right, cp4_0)
        #P1_left_mean,_ = map_point_direction_along_leg(P0_left_mean, None, leg, tapering)
        #P1_right_mean,_ = map_point_direction_along_leg(P0_right_mean, None, leg, tapering)
        #P1_left_left,_ = map_point_direction_along_leg(P0_left_left, None, leg, tapering)
        #P1_right_right,_ = map_point_direction_along_leg(P0_right_right, None, leg, tapering)
        #samples += [P1_left_mean, P1_right_mean, P1_left_left, P1_right_right]
        
        error_approx = half_hausdorff(cp4_approx, samples, level=5)
        
        if (not wild_handles(cp4_approx)) and (error_approx <= MAX_ERROR_ALLOWED):
            if print_info:
                print("OK")
                print("Append to finished_cp4list")
            finished_cp4list.append((cp4_approx, error_approx))
        elif subdiv_depth < MAX_LEVEL_ALLOWED:
            if print_info:
                print("NOT OK: Subdivide")
            subdiv = bezier_subdivide_at_parameters([t_mean], cp4_under_work)
            subdiv_with_level = [(sd, subdiv_depth+1) for sd in subdiv]
            unfinished_cp4list = subdiv_with_level + unfinished_cp4list
            subdiv_depth_max = max(subdiv_depth_max, subdiv_depth+1)
        else:
            if print_info:
                print("NOT OK")
                print("Put to finished since reached MAX_LEVEL_ALLOWED = "+str(MAX_LEVEL_ALLOWED))
            # Put to finished though not good.
            finished_cp4list.append((cp4_approx, error_approx))
    else:
        if print_info:
            print("No break")
    if len(finished_cp4list) > 0:
        cp4list = [cp4 for cp4,_ in finished_cp4list]
        total_error = max(err for _,err in finished_cp4list)
        if print_info:
            print("cp4_to_cp4list_distort: Returning cp4list:")
            for i,cp4 in enumerate(cp4list):
                print(str(i))
                for cp in cp4:
                    print(str(cp))
        return cp4list, total_error
    else:
        if print_info:
            print("Returning empty cp4list")
        return [],0

# ---------------------
# bc_to_cp4list_distort
# ---------------------
# Copy the bc:BCurve.BezierCurve onto the stroke leg,
# suitably scaled and rotated and distorted.
# The points 0 and 1 are mapped onto the leg end points.
# Args:
# - bc:       BCurve.BezierCurve
# - leg:      StrokeLeg
# - tapering: float
# Returns:
# - [[complex,complex,complex,complex]] or [[complex]] (=[cp4])
# - float (an error measure)
def bc_to_cp4list_distort(bc,
                          leg,
                          tapering):
    from cmath import phase, polar
    ZERO = 1e-8
    if (len(bc.bezier_arcs) == 1) and (len(bc.bezier_arcs[0].cp4) == 1):
        # 1-anchor stroke
        z0 = bc.bezier_arcs[0].cp4[0]
        z1,_ = map_point_direction_along_leg(z0, None, leg, tapering)
        new_balist = [BCurve.BezierArc(cp4=[z1])]
        return BCurve.BezierCurve(bezier_arcs=new_balist), 0 # set error=0
    new_cp4list = []
    bas = completed_bezier_arcs(bc)
    total_error = 0
    for ba in bas:
        cp4list, ba_error = cp4_to_cp4list_distort(ba.cp4,  # Main call
                                                   leg,
                                                   tapering
                                                   )
        new_cp4list += cp4list
        total_error = max(total_error, ba_error)
    if len(new_cp4list) == 0:
        m = "Could not manage with the task."
        m += "\nPlease check that the input paths"
        m += "\n    - the Source path,"
        m += "\n    - the Base path,"
        m += "\n    - the path to follow"
        m += "\nwere chosen right."
        raise Exception(m)
    new_balist = [BCurve.BezierArc(cp4=cp4) for cp4 in new_cp4list]
    return BCurve.BezierCurve(bezier_arcs=new_balist), total_error

# ---------------------
# bc_to_cp4list_as_such
# ---------------------
# Copy the bc:BCurve.BezierCurve onto the line segment between 
# start_point and end_point suitably scaled and rotated but without any distortion.
# The point 0 and 1 are mapped onto start_point and end_point.
# Args:
# - bc:          BCurve.BezierCurve
# - start_point: complex
# - end_point:   complex
# Returns:
# - BCurve.BezierCurve
def bc_to_cp4list_as_such(bc, start_point, end_point):
    P,Q = start_point, end_point
    similitude01PQ = make_similitude(0,1,P,Q)
    if (len(bc.bezier_arcs) == 1) and (len(bc.bezier_arcs[0].cp4) == 1):
        # 1-anchor stroke
        new_cp4list = [[similitude01PQ(bc.bezier_arcs[0].cp4[0])]]
    else:
        bas = completed_bezier_arcs(bc)
        new_cp4list = [[similitude01PQ(z) for z in ba.cp4] for ba in bas]
    new_balist = [BCurve.BezierArc(cp4=cp4) for cp4 in new_cp4list]
    return BCurve.BezierCurve(bezier_arcs=new_balist)

# -----------
# bc_onto_leg
# -----------
# Copy the bc:BCurve.BezierCurve onto the stroke leg,
# suitably scaled and rotated and possibly distorted.
# The points 0 and 1 are mapped onto start_point and end_point.
# Args:
# - bc:         BCurve.BezierCurve
# - leg:        StrokeLeg
# - distortion: boolean
# - tapering:   float
# Returns:
# - BCurve.BezierCurve
# - float (an error measure)
def bc_onto_leg(bc, leg, distortion, tapering):
    if not distortion:
        return (bc_to_cp4list_as_such(bc, leg.start, leg.end),
               0) # error=0
    else:
        return bc_to_cp4list_distort(bc, leg, tapering)

# ------------
# gv_onto_legs
# ------------
# Copy the gv (of the Source path in the standard position) onto the given
# list of legs, suitably scaled and rotated and possibly distorted.
# Distorted means that straight edges of the Source are arched according
# to path curvature.
# Args:
# - image
# - gv:         BCurve.GimpVectors
# - leglist:    [StrokeLeg]
# - distortion: boolean
# - tapering:   float
# Returns: 
# - [BCurve.GimpVectors]
def gv_onto_legs(image,
                 gv,         # Source
                 leglist,
                 distortion,
                 tapering
                 ):
    ZERO = 1e-6
    bclist = [gs.gs2bc() for gs in gv.stroke_list] # One bc for each stroke
    if distortion:
        bclist = preprocess_bclist(bclist) # Currently subdivides at inflection points
    #total_error = 0
    new_gvlist = []
    # Map the gv on each leg:
    for leg in leglist:
        # Map each bc of the Source to the leg:
        new_bclist = []
        for bc in bclist: # Source
            bc_on_leg, bc_error = bc_onto_leg(bc,             # Main call
                                              leg,
                                              distortion,
                                              tapering
                                              )
            if bc.closed:
                new_balist = bc_on_leg.bezier_arcs
                gap_ba = new_balist[-1]
                new_balist = new_balist[:-1]
                new_bc = BCurve.BezierCurve(bezier_arcs=new_balist,
                                            head_handle=gap_ba.cp4[2],
                                            tail_handle=gap_ba.cp4[1],
                                            closed=True
                                            )
            else:
                new_bc = bc_on_leg
            new_bclist.append(new_bc)
            #total_error = max(total_error, bc_error)
        # From the mapped bc's build new gv:
        new_gslist = [bc.bc2gs() for bc in new_bclist]
        new_gvlist.append(BCurve.GimpVectors(stroke_list=new_gslist))
    return new_gvlist


#==============================================================
#             Main procedure
#==============================================================

#distort_options = [ # (description, identifier)
#        ('distortion',       'distort'),
#        ('no distortion',    'no distortion'),
#        ]

merge_options = [ # (description, identifier)
        ('Merge all copies to one path',                'merge all'),
        ('Merge all copies on one stroke to one path',  'merge strokes'),
        ('Keep each copy as a separate path',           'merge nothing'),
        ]

MAX_NCOPIES = 1000

def replicate_path_along_path_main(image,
                                   source_path,
                                   base_path,
                                   target_path,
                                   leg_length,
                                   ncopies,
                                   tapering,     # percent, positive or negative
                                   distortion,   # boolean
                                   merge_option, # integer
                                   reverse_target
                                   ):
    MAX_TAPERING = 90
    ZERO = 1e-10
    tapering = min(tapering, MAX_TAPERING)   # case tapering >= 0
    tapering = -min(-tapering, MAX_TAPERING) # case tapering < 0
    tapering /= 100                          # No longer percents
    merge_case = merge_options[merge_option][1]
    ncopies = min(max(0, ncopies), MAX_NCOPIES)
    
    if len(base_path.strokes) != 1:
        raise Exception("The base path must have one stroke exactly")
    if len(base_path.strokes[0].points[0]) <= 6:
        raise Exception(("The base path must have at least two anchors in its only stroke."
                         +"\nThe end anchors will be used as the reference points."
                          ))
    if base_path.strokes[0].points[1]: # closed
        raise Exception("The base path must have an open stroke as its only stroke")
    
    gv_base = vectors_object2gv(base_path)
    base_ends = [gv_base.stroke_list[0].cp_list[1], # First and last anchors
                 gv_base.stroke_list[0].cp_list[-2]]
    if abs(base_ends[0]-base_ends[1]) < ZERO:
        raise Exception("The ends of the base path are too close to each other.")
    if abs(base_ends[0].real - base_ends[1].real) < ZERO: # Base upright
        if base_ends[0].imag < base_ends[1].imag: # Make the base from bottom to up
            base_ends.reverse()
            gv_base.stroke_list[0].cp_list.reverse()
    elif base_ends[0].real > base_ends[1].real: # Make the base from left to right
        base_ends.reverse()
        gv_base.stroke_list[0].cp_list.reverse()
    bc_base = gv_base.stroke_list[0].gs2bc()
    
    gv_source = vectors_object2gv(source_path)
    
    # Instead of using the input Base path, we imagine it transformed to
    # standard position by mapping the base_ends to points 0 and 1 in the
    # complex plane. We map the Source path with the same transformation:
    A,B = base_ends
    similitude_standard = make_similitude(A,B,0,1)
    strokes_source_standard = []
    for gs in gv_source.stroke_list:
        cp_list_standard = [similitude_standard(cp) for cp in gs.cp_list]
        strokes_source_standard.append(BCurve.GimpStroke(cp_list = cp_list_standard,
                                                          closed = gs.closed))
    gv_source_standard = BCurve.GimpVectors(stroke_list = strokes_source_standard)
    #
    # See where the extreme left and right points of gv_source_standard are
    # with respect to 0,1:
    NW,SE = gv_BB(gv_source_standard, Gimp=True)
    left_extreme = NW.real
    right_extreme = SE.real
    if reverse_target:
        gv_target = vectors_object2gv(target_path)
        for gs_target in gv_target.stroke_list:
            gs_target.cp_list.reverse()
        target_path = gv_target.gv2vectors_object(image, name=target_path.name+' reversed')
    
    # Preprocess target_path (make preprocessed copy).
    target_path = preprocess_target_path(image, target_path)
    
    # Create leglists along the target_path:
    delta = 0.1 # Just to be on the safer side we exaggerate the overshoots a little
    fill_strokes = (leg_length == 0)
    if (leg_length == 0) and (ncopies == 0):
        raise Exception(("It is not allowed to have the number of copies"
                        +"\nand the length of copies simultaneously zero."))
    elif fill_strokes:
        # Create leglists, one list for each stroke, filling each stroke
        # from start to end (taking into account ncopies but no leg_length):
        leglist_list = strokes_to_legs_by_ncopies(image,
                                  target_path,
                                  ncopies,
                                  -left_extreme + delta,     # left overshoot >= 0
                                  right_extreme + delta - 1, # right overshoot >= 0
                                  tapering)
    else:
        # Create leglists, one list for each stroke, starting with
        # the given leg_length:
        leglist_list = strokes_to_legs_by_length(image,
                                  target_path,
                                  leg_length,
                                  ncopies,
                                  -left_extreme + delta,     # left overshoot >= 0
                                  right_extreme + delta - 1, # right overshoot >= 0
                                  tapering)
    # For debugging:
    #print("Legs:")
    #for i,leglist in enumerate(leglist_list):
    #    print("List for stroke "+str(i))
    #    for j,leg in enumerate(leglist):
    #        print("leg "+str(j))
    #        print(str(leg))
    
    # Create transformed paths, one from each stroke:
    new_gvlist_list = []
    for stroke_index, target_stroke in enumerate(target_path.strokes):
        leglist = leglist_list[stroke_index] # Legs on this target_stroke
        if fill_strokes:
            work_legs = leglist
        elif ncopies == 0:
            work_legs = leglist
        else:
            work_legs = leglist[:ncopies] # Should be this already
        new_gvlist = gv_onto_legs(image,                    # Main call
                                  gv_source_standard,
                                  work_legs,
                                  distortion,
                                  tapering
                                  )
        new_gvlist_list.append(new_gvlist)
    # Do mergings and re-build new_gvlist:
    if merge_case == 'merge all':
        one_gs_list = [] # All gs's in one list
        for gvlist in new_gvlist_list:
            for gv in gvlist:
                one_gs_list += gv.stroke_list
        one_gv = BCurve.GimpVectors(stroke_list = one_gs_list)
        one_gv.name = source_path.name+'|replicate'
        new_gvlist_list = [[one_gv]]
    elif merge_case == 'merge strokes':
        stroke_gvs = []
        stroke_count = 0
        for gvlist in new_gvlist_list:
            gs_list = [] # All gs's in this stroke
            for gv in gvlist:
                gs_list += gv.stroke_list
            stroke_gv = BCurve.GimpVectors(stroke_list = gs_list)
            stroke_gv.name = source_path.name+'|replicate on stroke '+str(stroke_count)
            stroke_gvs.append([stroke_gv])
            stroke_count += 1
        new_gvlist_list = stroke_gvs
    elif merge_case == 'merge nothing': # Nothing to do except to set the names
        replica_count = 0
        for gvlist in new_gvlist_list:
            for gv in gvlist:
                gv.name = source_path.name+'|replica '+str(replica_count)
                replica_count += 1
    # Draw
    pdb.gimp_image_undo_group_start(image)
    for gvlist in new_gvlist_list:
        for gv in gvlist:
            new_path = gv.gv2vectors_object(image)
            gimp_draw_vectors_object(image, new_path, visible=True)
    ## For debugging:
    #extended_paths = [leglist[0].path for path in leglist_list]
    #for ext_path in extended_paths:
    #    gimp_draw_vectors_object(image, ext_path, visible=True)
    pdb.gimp_image_undo_group_end(image)
    return new_path

#======================================================
#                    Registration
#======================================================

versionnumber = "0.8"
procedure_author = "Markku Koppinen"
procedure_copyright = procedure_author
procedure_date = "2022"
image_types = "*"
menupath = '<Vectors>/Tools/Replicate'

procedure_name  = "replicate_path_along_path"
procedure_blurb = ("Replicate a path along another path,"
                   +"\noptionally distorted to follow the arc."
                   +"\nWorks best when the target path"
                   +"\nis smooth and only gently curving."
                   +"\nWarning: The result may need manual corrections."
                   +"\n(Version "+versionnumber+")"
                   )
procedure_help  = ("Assume given three paths: (1) the Source path;"
                   +"\n(2) the Base path (a 2-anchor line segment);"
                   +"\n(3) the Target path (the path to follow)."
                   +"\nThe Source path is replicated along each stroke"
                   +"\nof the Target path by transformations which send"
                   +"\nthe end points of the Base path onto the Target path."
                   +"\nAs a result, a succession of copies of the Source path"
                   +"\nis made along the Target path."
                   +"\nThe transformations rotate and scale and optionally"
                   "\ndistort the Source path to follow the Target path."
                   +"\nOptional tapering: successive copies of the Source path"
                   +"\nare scaled down (or up) by a fixed factor."
                   +"\nWorks well only when the Target path is smooth and"
                   +"\njust gently curving, without any too tight bends."
                   )

procedure_label = "Replicate path along another path"

procedure_function = replicate_path_along_path_main

register(
    procedure_name,
    procedure_blurb,
    procedure_help,
    procedure_author,
    procedure_copyright,
    procedure_date,
    procedure_label,
    image_types,
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_VECTORS, "source_path", "The Source path: the path to replicate", None),
      (PF_VECTORS, "base_path", 
                         ("The Base path: a 2-anchor line segment"
                         +"\n  The end points will serve as the reference"
                         +"\n  points to be positioned on the Target path."
                         ), None),
      (PF_VECTORS, "target_path", "The Target path: the path to follow", None),
      (PF_FLOAT, "leg_length", ("The length of copies (if tapering: of the first copy)"
                               +"\n  (Setting this to 0 means:"
                               +"\n  fill each stroke from start to end.)"
                               ), 100),
      (PF_INT,   "ncopies", ("Number of copies (0.."+str(MAX_NCOPIES)+")"
                            +"\n  (Setting this to 0 means:"
                            +"\n  make as many copies as there is room.)"
                            ), 3),
      (PF_FLOAT, "tapering", "Tapering (percents, -90..90)", 0.),
      (PF_BOOL, "distortion", "Distort the copies to follow the Target path?", True),
      (PF_OPTION, 'merge_option','Merge the copies?',
                   0,
                  [case[0] for case in merge_options]), # Descriptions of cases
      (PF_BOOL, "reverse_target", "Use reversed version of the Target path?", False),
    ],
    [
        (PF_VECTORS, "new_path", "New path"),
    ],
    procedure_function,
    menu=menupath)

main()
